#!/usr/bin/env python

import os
from unittest import mock

import pytest
from boutiques_mocks import example_boutiques_tool, mock_get

import boutiques as bosh
from boutiques import __file__ as bfile
from boutiques.tests.BaseTest import BaseTest


class TestEvaluate(BaseTest):
    @pytest.fixture(autouse=True)
    def set_test_dir(self):
        self.setup(
            os.path.join(os.path.dirname(bfile), "schema", "examples", "example1")
        )

    def test_evaloutput(self):
        desc = self.example1_descriptor
        invo = self.get_file_path("invocation.json")
        query = bosh.evaluate(desc, invo, "output-files/")
        expect = {
            "logfile": "./test_temp/log-4-coin;plop.txt",
            "output_files": "output/*_exampleOutputTag.resultType",
            "config_file": "./test_temp/config.txt",
        }
        self.assertEqual(query, expect)

        query = bosh.evaluate(desc, invo, "output-files/id=logfile")
        expect = {"logfile": "./test_temp/log-4-coin;plop.txt"}
        self.assertEqual(query, expect)

        query = bosh.evaluate(desc, invo, "output-files/id=log-file")
        expect = {}
        self.assertEqual(query, expect)

    def test_evaloutput_json_obj(self):
        desc = open(self.example1_descriptor).read()
        invo = open(self.get_file_path("invocation.json")).read()
        query = bosh.evaluate(desc, invo, "output-files/")
        expect = {
            "logfile": "./test_temp/log-4-coin;plop.txt",
            "output_files": "output/*_exampleOutputTag.resultType",
            "config_file": "./test_temp/config.txt",
        }
        self.assertEqual(query, expect)

        query = bosh.evaluate(desc, invo, "output-files/id=logfile")
        expect = {"logfile": "./test_temp/log-4-coin;plop.txt"}
        self.assertEqual(query, expect)

        query = bosh.evaluate(desc, invo, "output-files/id=log-file")
        expect = {}
        self.assertEqual(query, expect)

    @mock.patch("requests.get", return_value=mock_get())
    def test_evaloutput_from_zenodo(self, _):
        desc = "zenodo." + str(example_boutiques_tool.id)
        invo = self.get_file_path("invocation.json")
        query = bosh.evaluate(desc, invo, "output-files/")
        expect = {
            "logfile": "./test_temp/log-4-coin;plop.txt",
            "output_files": "output/*_exampleOutputTag.resultType",
            "config_file": "./test_temp/config.txt",
        }
        self.assertEqual(query, expect)

        query = bosh.evaluate(desc, invo, "output-files/id=logfile")
        expect = {"logfile": "./test_temp/log-4-coin;plop.txt"}
        self.assertEqual(query, expect)

        query = bosh.evaluate(desc, invo, "output-files/id=log-file")
        expect = {}
        self.assertEqual(query, expect)

    def test_evalinput(self):
        desc = self.example1_descriptor
        invo = self.get_file_path("invocation.json")
        query = bosh.evaluate(desc, invo, "inputs/")
        expect = {
            "str_input_list": ["fo '; echo FAIL", "bar"],
            "str_input": "coin;plop",
            "config_num": 4,
            "num_input": None,
            "file_input": "./pyproject.toml",
            "file_list_input": ["/test_mount1", "/test_mount2"],
            "enum_input": "val1",
            "list_int_input": [1, 2, 3],
            "flag_input": None,
            "no_opts": None,
        }
        self.assertEqual(query, expect)

        query = bosh.evaluate(
            desc, invo, "inputs/type=Flag,id=flag_input", "inputs/type=Number"
        )
        expect = [
            {"flag_input": None},
            {"config_num": 4, "num_input": None, "list_int_input": [1, 2, 3]},
        ]
        self.assertEqual(query, expect)

        query = bosh.evaluate(desc, invo, "inputs/id=strinputs")
        expect = {}
        self.assertEqual(query, expect)

        query = bosh.evaluate(desc, invo, "inputt/nonsense=strinputs")
        expect = {}
        self.assertEqual(query, expect)

    def test_evalgroups(self):
        desc = self.example1_descriptor
        invo = self.get_file_path("invocation.json")
        query = bosh.evaluate(desc, invo, "groups/")
        expect = {"an_example_group": {"num_input": None, "enum_input": "val1"}}
        self.assertEqual(query, expect)

        query = bosh.evaluate(desc, invo, "groups/mutually-exclusive=True")
        expect = {"an_example_group": {"num_input": None, "enum_input": "val1"}}
        self.assertEqual(query, expect)



#!/usr/bin/env python

import subprocess
from unittest import mock

import pytest
from boutiques_mocks import example_boutiques_tool, mock_zenodo_search
from jsonschema.exceptions import ValidationError

from boutiques import bosh
from boutiques.tests.BaseTest import BaseTest


class TestTest(BaseTest):
    @pytest.fixture(autouse=True)
    def set_test_dir(self):
        self.setup("test")

    @pytest.mark.skipif(
        subprocess.Popen("type docker", shell=True).wait(),
        reason="Docker not installed",
    )
    def test_test_good(self):
        self.assertFalse(bosh(["test", self.get_file_path("tests_good.json")]))

    @pytest.mark.skipif(
        subprocess.Popen("type docker", shell=True).wait(),
        reason="Docker not installed",
    )
    def test_test_good_desc_as_json_obj(self):
        self.assertFalse(
            bosh(["test", open(self.get_file_path("tests_good.json")).read()])
        )

    @pytest.mark.skipif(
        subprocess.Popen("type docker", shell=True).wait(),
        reason="Docker not installed",
    )
    @mock.patch(
        "requests.get",
        return_value=mock_zenodo_search([example_boutiques_tool]),
    )
    def test_test_good_from_zenodo(self, mock_get):
        self.assertFalse(bosh(["test", "zenodo." + str(example_boutiques_tool.id)]))

    def test_test_invalid(self):
        with self.assertRaises(ValidationError) as context:
            bosh(["test", self.get_file_path("tests_invalid.json")])
        error_1 = (
            'TestError: "this_id_does_not_exist" output id not '
            'found, in test "test1"'
        )
        error_2 = (
            'TestError: "logfile" output id cannot appear more'
            ' than once within same test, in test "test1"'
        )
        error_3 = 'TestError: "testNameDefinedTwice" test name is non-unique'
        self.assertTrue(error_1 in context.exception.message)
        self.assertTrue(error_2 in context.exception.message)
        self.assertTrue(error_3 in context.exception.message)
        self.assertTrue(context.exception.message.count("TestError") == 3)

    # Each of the following tests verify that pytest return with
    # an exit-code of '1'.
    # According to the pytest documentation, an exit-code of '1'
    # indicate that some _and not all_ of the tests have failed.
    # However, as each of the descriptors used for the following tests
    # define only one test, we can safely assume that such an
    # exit-code would indicate complete test failure.

    @pytest.mark.skipif(
        subprocess.Popen("type docker", shell=True).wait(),
        reason="Docker not installed",
    )
    def test_test_failure_mismatched_exitcode(self):
        self.assertEqual(
            1,
            bosh(["test", self.get_file_path("tests_failure_exitcode.json")]),
        )

    @pytest.mark.skipif(
        subprocess.Popen("type docker", shell=True).wait(),
        reason="Docker not installed",
    )
    def test_test_failure_mismatched_reference_content(self):
        self.assertEqual(
            1,
            bosh(["test", self.get_file_path("tests_failure_reference.json")]),
        )

    @pytest.mark.skipif(
        subprocess.Popen("type docker", shell=True).wait(),
        reason="Docker not installed",
    )
    def test_test_failure_unproduced_output(self):
        self.assertEqual(
            1,
            bosh(["test", self.get_file_path("tests_failure_output_id.json")]),
        )



#!/usr/bin/env python

import datetime
import hashlib
import math
import os
import os.path as op
import random
import random as rnd
import re
import string
import subprocess
import sys
import time
from glob import glob
from pathlib import Path

import simplejson as json
from termcolor import colored

import boutiques
from boutiques.dataHandler import getDataCacheDir
from boutiques.evaluate import evaluateEngine
from boutiques.logger import print_info, print_warning, raise_error
from boutiques.util.utils import conditionalExpFormat, extractFileName, loadJson


class ExecutorOutput:
    def __init__(
        self,
        stdout,
        stderr,
        exit_code,
        desc_err,
        output_files,
        missing_files,
        shell_command,
        container_command,
        container_location,
    ):
        try:
            self.stdout = stdout.decode("utf=8", "backslashreplace")
        except AttributeError as e:
            self.stdout = stdout
        try:
            self.stderr = stderr.decode("utf=8", "backslashreplace")
        except AttributeError as e:
            self.stderr = stderr
        self.exit_code = exit_code
        self.error_message = desc_err
        self.output_files = output_files
        self.missing_files = missing_files
        self.shell_command = shell_command
        self.container_command = container_command
        self.container_location = container_location

    def __str__(self):
        formatted_output_files = ""
        for f in self.output_files:
            if formatted_output_files != "":
                formatted_output_files += os.linesep
            formatted_output_files += "\t- " + str(f)

        formatted_missing_files = ""
        for f in self.missing_files:
            if formatted_missing_files != "":
                formatted_missing_files += os.linesep
            formatted_missing_files += "\t- " + str(f)

        def title(s):
            return colored(s + os.linesep, "green")

        out = (
            title("Shell command")
            + "{0}"
            + os.linesep
            + title("Container location")
            + "{1}"
            + os.linesep
            + title("Container command")
            + "{2}"
            + os.linesep
            + title("Exit code")
            + "{3}"
            + os.linesep
            + (title("Std out") + "{4}" + os.linesep if self.stdout else "")
            + (
                title("Std err") + colored("{5}", "red") + os.linesep
                if self.stderr
                else ""
            )
            + title("Error message")
            + colored("{6}", "red")
            + os.linesep
            + title("Output files")
            + "{7}"
            + os.linesep
            + title("Missing files")
            + colored("{8}", "red")
            + os.linesep
        ).format(
            self.shell_command,
            self.container_location,
            self.container_command,
            self.exit_code,
            self.stdout,
            self.stderr,
            self.error_message,
            formatted_output_files,
            formatted_missing_files,
        )
        return out


class FileDescription:
    def __init__(self, boutiques_name, file_name, optional):
        self.boutiques_name = boutiques_name
        self.file_name = file_name
        self.optional = "Optional"
        if not optional:
            self.optional = "Required"

    def __str__(self):
        return f"{self.file_name} ({self.boutiques_name}, {self.optional})"


class ExecutorError(Exception):
    pass


# Executor class
class LocalExecutor:
    """
    This class represents a json descriptor of a tool, and can execute
    various tasks related to it. It is constructed first via an
    input json descriptor file, which is held in the desc_dict field.

    An input can be added to it via the in_dict field, a dictionary from
    param ids to values. The in_dict field should only be modified
    via the public readInput method, which can either take
    a file (json or csv) or a string written in the command line.
    The field is always validated by checking the input parameters
     with respect to the descriptor.

    Other public methods include:
        execute - attempts to execute the tool described by the descriptor
            based on the current input (in in_dict)
        printCmdLine - simply prints the generated command line
            based on the current input values
        generateRandomParams - fills in_dict with random values
            (following the constraints of the descriptor schema)
    """

    # Constructor
    def __init__(self, desc, invocation, options={}):
        self._rkit = self._replaceKeysInTemplate  # Abbrev. for readability

        # Initial parameters
        self.desc_path = desc  # Save descriptor path
        self.errs = []  # Empty errors holder
        self.invocation = invocation

        # Extra Options
        # Include: forcePathType and debug
        self.debug = False
        for option in list(options.keys()):
            setattr(self, option, options.get(option))

        # Parse JSON descriptor
        self.desc_dict = loadJson(desc, self.debug, sandbox=self.sandbox)

        # Set the shell
        self.shell = self.desc_dict.get("shell")
        if self.shell is None:
            self.shell = "/bin/sh"

        # Generate Summary for data collection
        if not self.skipDataCollect:
            self.summary = self._generateSummary(desc)

        # Helpers Functions
        # The set of input parameters from the json descriptor
        self.inputs = self.desc_dict["inputs"]  # Struct: [{id:}..,{id:}]
        # The set of output parameters from the json descriptor
        self.outputs = self.desc_dict.get("output-files") or []
        # The set of parameter groups, according to the json descriptor
        self.groups = self.desc_dict.get("groups") or []

        # Container-image Options
        self.con = self.desc_dict.get("container-image")
        self.launchDir = None
        if self.con is not None:
            self.con.get("working-directory")

        # Generate the command line
        if self.invocation:
            self.readInput(self.invocation)

    # Retrieves the parameter corresponding to the given id
    def byId(self, n):
        return [v for v in self.inputs + self.outputs if v["id"] == n][0]

    # Retrieves the group corresponding to the given id
    def byGid(self, g):
        return [v for v in self.groups if v["id"] == g][0]

    # Retrieves the value of a field of an input
    # from the descriptor. Returns None if not present.
    def safeGet(self, i, k):
        if k not in list(self.byId(i).keys()):
            return None
        return self.byId(i)[k]

    # Retrieves the value of a field of a group from
    # the descriptor. Returns None if not present.
    def safeGrpGet(self, g, k):
        if k not in list(self.byGid(g).keys()):
            return None
        return self.byGid(g)[k]

    # Retrieves the group a given parameter id belongs to;
    # otherwise, returns None
    def assocGrp(self, i):
        return ([g for g in self.groups if i in g["members"]] or [None])[0]

    # Returns the required inputs of a given input id, or the empty string
    def reqsOf(self, t):
        if t in [g["id"] for g in self.groups]:
            return self.safeGrpGet(t, "members") or []
        return self.safeGet(t, "requires-inputs") or []

    # Attempt local execution of the command line
    # generated from the input values
    def execute(self, mount_strings, conOpts=None):
        """
        The execute method runs the generated command line
        (from either generateRandomParams or readInput)
        If docker is specified, it will attempt to use it, instead
        of local execution.
        After execution, it checks for output file existence.
        """
        command, exit_code, con = self.cmd_line[0], None, self.con or {}
        # Check for Container image
        conType, conImage = (
            con.get("type"),
            (con.get("image") if not self.noContainer else None),
        )
        conIndex = con.get("index")
        # conOpts can't be forced for docker, as long as there are cases
        # where _chooseContainerTypeToUse() can fall back to singularity
        conOptsForced = conOpts is not None and self.forceSingularity
        conOpts = conOpts or con.get("container-opts")
        conIsPresent = conImage is not None
        # Export environment variables,
        #  if they are specified in the descriptor
        envVars = {}
        if "environment-variables" in list(self.desc_dict.keys()):
            variables = [
                (p["name"], p["value"]) for p in self.desc_dict["environment-variables"]
            ]
            inputsByValKey = {inp["value-key"]: inp for inp in self.inputs}
            for envVarName, envVarValue in variables:
                if envVarValue in inputsByValKey:
                    envVarValue = self.in_dict[inputsByValKey[envVarValue]["id"]]
                os.environ[envVarName] = envVarValue
                envVars[envVarName] = envVarValue
        # Container script constant name
        # Note that docker/singularity cannot do a local volume
        # mount of files starting with a '.', hence this one does not
        millitime = int(time.time() * 1000)
        dsname = (
            "temp-"
            + str(random.SystemRandom().randint(0, int(millitime)))
            + "-"
            + str(millitime)
            + ".localExec.boshjob.sh"
        )
        dsname = op.realpath(dsname)
        # If container is present, alter the command template accordingly
        container_location = ""
        container_command = ""
        if conIsPresent:
            # Figure out which container type to use
            conTypeToUse = self._chooseContainerTypeToUse(
                conType, self.forceSingularity, self.forceDocker
            )
            # Pull the container
            (conPath, container_location) = self.prepare(conTypeToUse)
            # Generate command script
            # Get the supported shell by the docker or singularity
            cmdString = f"#!{self.shell}"
            if self.shell == "/bin/sh":
                cmdString += " -l"
            cmdString += os.linesep + str(command)
            with open(dsname, "w") as scrFile:
                scrFile.write(cmdString)
            # Ensure the script is executable
            self._localExecute("chmod 755 " + dsname)
            # Prepare extra environment variables
            envString = ""
            if envVars:
                for key, val in list(envVars.items()):
                    envString += f"SINGULARITYENV_{key}='{val}' "
            # Change launch (working) directory if desired
            launchDir = self.launchDir
            if launchDir is None:
                launchDir = op.realpath("./")
            launchDir = op.realpath(launchDir)
            # Get the container options
            conOptsString = ""
            if conOpts:
                # Ignore container options if container type is not the one
                # specified in the descriptor.
                if conType != conTypeToUse and not conOptsForced:
                    print_warning("Ignoring incompatible container options.")
                else:
                    for opt in conOpts:
                        conOptsString += opt + " "
            # Run it in docker
            # Note: on Windows, users must give path following
            #       this format (compatible with docker): /c/a/windows/path
            mount_strings = [] if not mount_strings else mount_strings

            # Normalize the path so that it follows
            #  this docker compatible format: /c/a/windows/or/linux/path
            # Do nothing on linux paths
            # If the path begins with C: or any other capital letter:
            #  - replace '\\' with '/'
            #  - prefix the path with '/'
            #  - lowercase the drive letter
            #  - remove the ':'
            def normalizePath(path):
                regexResult = re.match(r"^([A-Z]):", path)
                if regexResult:
                    path = path.replace("\\", "/")
                    path = "/" + path[0].lower() + path[2:]
                return path

            # Make path absolute and normalized
            # The resulting path must follow
            # this docker compatible format: /c/a/windows/or/linux/path
            def makePathAbsolute(path):
                # If path is already absolute: do nothing
                # (Note that on Windows, op.realpath(/c/path/to/file)
                #  returns C:\\c\\path\\to\\file, so we should
                #  avoid applying op.realpath() if already absolute)
                # (On both Windows and Linux,
                #  paths beginning with '/' are considered absolute)
                if op.isabs(path):
                    # If path is absolute, it must be normalized
                    return normalizePath(path)
                # Make path absolute
                path = op.realpath(path)
                # Normalize it
                return normalizePath(path)

            def addAutomounts(mount_strings, launchDir):
                # Extend list of mounts with all files in invocation
                mount_inputs = []
                existing_mounts = [m.split(":")[1] for m in mount_strings]
                for file_input in [
                    i for i in self.inputs if i["type"].lower() == "file"
                ]:
                    if file_input["id"] in self.in_dict:
                        if "list" in file_input and file_input["list"]:
                            mount_inputs.extend(self.in_dict[file_input["id"]])
                        else:
                            mount_inputs.append(self.in_dict[file_input["id"]])

                # Prevent duplicate mounts,
                mount_inputs = [m for m in mount_inputs if m not in existing_mounts]

                # Ensure the mount location exist
                missing_mounts = []
                for m in mount_inputs:
                    abs_path = makePathAbsolute(m)
                    if not Path(abs_path).exists():
                        missing_mounts.append(abs_path)
                if missing_mounts:
                    raise_error(
                        ExecutorError,
                        f"Missing local mount location: {missing_mounts}",
                    )

                # normalize paths and add file under same absolute path
                mount_inputs = [
                    makePathAbsolute(m) + ":" + Path(launchDir, m).resolve().as_posix()
                    for m in mount_inputs
                ]
                mount_strings.extend(mount_inputs)
                return mount_strings

            launchDir = normalizePath(launchDir)
            dsname = normalizePath(dsname)

            mount_strings = [
                makePathAbsolute(m.split(":")[0]) + ":" + m.split(":")[1]
                for m in mount_strings
            ]
            mount_strings.append(makePathAbsolute("./") + ":" + launchDir)

            if not self.noAutomounts:
                mount_strings = addAutomounts(mount_strings, launchDir)

            if conTypeToUse == "docker":
                envString = " "
                if envVars:
                    for key, val in list(envVars.items()):
                        envString += f" -e {key}='{val}' "
                # export mounts to docker string
                docker_mounts = " -v ".join(m for m in mount_strings)
                # If --changeUser was desired, provides the current user id
                # and its group id as the user and group to be used instead
                # of the default root within the container.
                userchange = ""
                if self.changeUser:
                    userchange = " -u $(id -u):$(id -g)"

                container_command = (
                    "docker run"
                    + userchange
                    + " --entrypoint="
                    + self.shell
                    + " --rm"
                    + envString
                    + " -v "
                    + docker_mounts
                    + " -w "
                    + launchDir
                    + " "
                    + conOptsString
                    + str(conImage)
                    + " "
                    + dsname
                )
            elif conTypeToUse == "singularity":
                envString = ""
                if envVars:
                    for key, val in list(envVars.items()):
                        envString += f"SINGULARITYENV_{key}='{val}' "
                singularity_mounts = "-B " + " -B ".join(mount_strings)
                container_command = (
                    envString + "singularity exec "
                    "--cleanenv "
                    + singularity_mounts
                    + " -W "
                    + launchDir
                    + " "
                    + conOptsString
                    + str(conPath)
                    + " "
                    + dsname
                )
            (stdout, stderr), exit_code = self._localExecute(container_command)
        # Otherwise, just run command locally
        else:
            (stdout, stderr), exit_code = self._localExecute(command)
        time.sleep(0.5)  # Give the OS a (half) second to finish writing

        # Destroy temporary docker script, if desired.
        # By default, keep the script so the dev can look at it.
        if conIsPresent and not self.debug:
            if os.path.isfile(dsname):
                os.remove(dsname)

        # Check for output files
        missing_files = []
        missing_files_dict = {}
        output_files = []
        output_files_dict = {}
        if "output-files" in list(self.desc_dict.keys()):
            all_files = evaluateEngine(self, "output-files")
            required_files = evaluateEngine(self, "output-files/optional=False")
            optional_files = evaluateEngine(self, "output-files/optional=True")
            for f in all_files.keys():
                file_name = all_files[f]
                fd = FileDescription(f, file_name, False)
                f_glob = glob(file_name)
                if f_glob:
                    fd.file_name = f_glob[0]
                    output_files.append(fd)
                    output_files_dict[f] = f_glob[0]
                else:  # file does not exist
                    if f in required_files.keys():
                        missing_files.append(fd)
                        missing_files_dict[f] = file_name

        # Set error messages
        desc_err = ""
        if "error-codes" in list(self.desc_dict.keys()):
            for err_elem in self.desc_dict["error-codes"]:
                if err_elem["code"] == exit_code:
                    desc_err = err_elem["description"]
                    break

        executor_output = ExecutorOutput(
            stdout,
            stderr,
            exit_code,
            desc_err,
            output_files,
            missing_files,
            command,
            container_command,
            container_location,
        )

        if not self.skipDataCollect:
            # Generate public output
            self.public_out = self._generatePublicOutput(
                executor_output, output_files_dict, missing_files_dict
            )
            # Write data collection to file
            self._saveDataCaptureToCache()

        return executor_output

    # Looks for the container image locally and pulls it if not found
    # Returns a tuple containing the container filename (for Singularity)
    # and the container location (local or pulled)
    def prepare(self, conTypeToUse=None):
        con = self.con
        if con is None:
            return ("", "Descriptor does not specify a container image.")

        conType, conImage = (
            con.get("type"),
            con.get("image"),
        )
        conIndex = con.get("index")

        # If container is present, alter the command template accordingly
        conName = ""

        if conTypeToUse is None:
            conTypeToUse = self._chooseContainerTypeToUse(conType)

        if conTypeToUse == "docker":
            # Pull the docker image
            if self._localExecute("docker pull " + str(conImage))[1]:
                container_location = "Local copy"
            else:
                container_location = "Pulled from Docker"
            return (conName, container_location)

        elif conTypeToUse == "singularity":
            if conType == "docker":
                # We're running a Docker image in Singularity
                conIndex = "docker://" + (
                    conIndex
                    if (
                        conIndex is not None
                        and conIndex != ""
                        and conIndex != "docker://"
                    )
                    else ""
                )

            # If present, assign index in conImage to conIndex
            if re.search(r"^[a-zA-Z0-9]+://", conImage) is not None:
                conIndex = re.match(r"^[a-zA-Z0-9]+://", conImage).group()
                conImage = conImage.replace(conIndex, "")
            if not conIndex:
                conIndex = "shub://"
            if not conIndex.endswith("/"):
                conIndex = conIndex + "/"

            if self.imagePath:
                conName = op.basename(self.imagePath)
                imageDir = op.normpath(op.dirname(self.imagePath))
            else:
                conName = conImage.replace("/", "-").replace(":", "-") + ".simg"
                imageDir = op.normpath("")

            # Check if container already exists
            if self._singConExists(conName, imageDir):
                conPath = op.abspath(op.join(imageDir, conName))
                return conPath, f"Local ({conName})"

            # Container does not exist, try to pull it
            if self.imagePath:
                lockDir = self.imagePath + "-lock"
            else:
                lockDir = conName + "-lock"

            maxcount = 36
            count = 0

            while count < maxcount:
                count += 1
                try:
                    os.mkdir(lockDir)
                except OSError:
                    print_info(
                        "Another process seems to be pulling the "
                        "image ({} exists), sleeping 5 seconds".format(lockDir)
                    )
                    time.sleep(5)
                else:
                    try:
                        # Check if container was created while waiting
                        if self._singConExists(conName, imageDir):
                            conPath = op.abspath(op.join(imageDir, conName))
                            container_location = f"Local ({conName})"
                        # Container still does not exist, so pull it
                        else:
                            conPath, container_location = self._pullSingImage(
                                conName, conIndex, conImage, imageDir, lockDir
                            )
                        return conPath, container_location
                    finally:
                        self._cleanUpAfterSingPull(lockDir)
            # If loop times out, check again for existence, otherwise
            # raise an error
            if self._singConExists(conName, imageDir):
                conPath = op.abspath(op.join(imageDir, conName))
                return conPath, f"Local ({conName})"
            raise_error(ExecutorError, "Unable to retrieve Singularity " "image.")

    # Private method that checks if a Singularity image exists locally
    def _singConExists(self, conName, imageDir):
        return conName in os.listdir(imageDir)

    # Private method that pulls a Singularity image
    def _pullSingImage(self, conName, conIndex, conImage, imageDir, lockDir):
        # Give the file a temporary name while it's building
        conNameTmp = conName + ".tmp"
        # Set the pull directory to the specified imagePath
        if self.imagePath:
            os.environ["SINGULARITY_PULLFOLDER"] = imageDir
        pull_loc = f'"{conNameTmp}" {conIndex}{conImage}'
        container_location = (
            "Pulled from {1}{2} ({0} not found "
            "in current working "
            "directory or specified "
            "image path)"
        ).format(conName, conIndex, conImage)
        # Pull the singularity image
        sing_command = "singularity pull --name " + pull_loc
        (stdout, stderr), return_code = self._localExecute(sing_command)
        if return_code:
            message = (
                "Could not pull Singularity"
                " image: "
                + os.linesep
                + " * Pull command: "
                + sing_command
                + os.linesep
                + " * Error: "
                + stderr.decode("utf-8")
            )
            raise_error(ExecutorError, message)
        os.rename(op.join(imageDir, conNameTmp), op.join(imageDir, conName))
        conPath = op.abspath(op.join(imageDir, conName))
        return conPath, container_location

    # Removes the lock directory and environment variable
    # created while pulling an image
    def _cleanUpAfterSingPull(self, lockDir):
        os.rmdir(lockDir)
        if "SINGULARITY_PULLFOLDER" in os.environ:
            del os.environ["SINGULARITY_PULLFOLDER"]

    def _isCommandInstalled(self, command):
        return not subprocess.Popen(f"{command} --version", shell=True).wait()

    # Chooses whether to use Docker or Singularity based on the
    # descriptor, executor options and if Docker is installed.
    def _chooseContainerTypeToUse(self, conType, forceSing=False, forceDocker=False):
        if (
            conType == "docker" and not forceSing or forceDocker
        ) and self._isCommandInstalled("docker"):
            return "docker"

        if self._isCommandInstalled("singularity"):
            return "singularity"

        raise_error(
            ExecutorError,
            (
                "Could not find any container engine. "
                + "Make sure that Docker or Singularity "
                + "is installed."
            ),
        )

    # Private method that attempts to locally execute the given
    # command. Returns the exit code.
    def _localExecute(self, command):
        # Note: invokes the command through the shell
        # (potential injection dangers)
        if self.debug:
            print_info(f"Running: {command}")
        try:
            if self.stream:
                process = subprocess.Popen(
                    command,
                    shell=True,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT,
                )
            else:
                process = subprocess.Popen(
                    command,
                    shell=True,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                )

        except OSError as e:
            sys.stderr.write("OS Error during attempted execution!")
            raise e
        except ValueError as e:
            sys.stderr.write("Input Value Error during attempted execution!")
            raise e

        if not self.stream:
            return process.communicate(), process.returncode

        while True:
            if process.poll() is not None:
                break
            outLine = process.stdout.readline().decode()
            if outLine != "":
                sys.stdout.write(outLine)
        # Return (stdout, stderr) as (None, None) since it was already
        # printed in real time
        return (None, None), process.returncode

    # Private method to generate a random input parameter set that follows
    # the constraints from the json descriptor
    # This method fills in the in_dict field of the object
    # with constrained random values
    def _randomFillInDict(self):
        # Private helper functions for filling the dictionary
        # Helpers for generating random numbers, strings, etc...
        # Note: uses-absolute-path is satisfied for files by the
        # automatic replacement in the _validateDict
        # nd = number of random characters to use in generating strings,
        # nl = max number of random list items
        nd, nl = 2, 5
        # number_decimals = the number of decimal places to round to
        # on randomly generated floating point numbers
        number_decimals = 3
        # epsilon = an upper bound on the relative error due to
        # rounding for the chosen number of decimals, making sure
        # excluded values aren't rounded to by accident.
        epsilon = 1.0 / (10**number_decimals)

        def randDigs():
            # Generate random string of digits
            return "".join(rnd.choice(string.digits) for _ in range(nd))

        def randFile(id):
            return (
                "f_"
                + id
                + "_"
                + randDigs()
                + rnd.choice(
                    [".csv", ".tex", ".j", ".cpp", ".m", ".mnc", ".nii.gz", ""]
                )
            )

        def randStr(id):
            return (
                "str_"
                + id
                + "_"
                + "".join(
                    rnd.choice(string.digits + string.ascii_letters) for _ in range(nd)
                )
            )

        # A function for generating a number type parameter input
        # p is a dictionary object corresponding to a parameter
        # description in the json descriptor
        # E.g. if p had no constraints from the json descriptor,
        # the output would be a float in [defaultMin,defaultMax]
        #            if p had "integer": true, "minimum": 7,
        # "maximum": 9, the output would be an int in [7,9]
        def randNum(p):
            param_id, defaultMin, defaultMax = p["id"], -50, 50
            # Check if the input parameter should be an int
            isInt = self.safeGet(param_id, "integer")

            def roundTowardsZero(x):
                return int(math.copysign(1, x) * int(abs(x)))

            # Assign random values to min and max,
            # unless they have been specified
            minv = self.safeGet(param_id, "minimum")
            maxv = self.safeGet(param_id, "maximum")
            if minv is None and maxv is None:
                minv, maxv = defaultMin, defaultMax
            elif minv is None and not (maxv is None):
                minv = maxv + defaultMin
            elif not (minv is None) and maxv is None:
                maxv = minv + defaultMax
            # Coerce the min/max to the proper number type
            if isInt:
                minv, maxv = roundTowardsZero(minv), roundTowardsZero(maxv)
            else:
                minv, maxv = float(minv), float(maxv)
            # Apply exclusive boundary constraints, if any
            if self.safeGet(param_id, "exclusive-minimum"):
                minv += 1 if isInt else epsilon
            if self.safeGet(param_id, "exclusive-maximum"):
                maxv -= 1 if isInt else epsilon
            # Returns a random int or a random float, depending on the type of p
            return (
                rnd.randint(minv, maxv)
                if isInt
                else round(rnd.uniform(minv, maxv), number_decimals)
            )

        # Generate a random parameter value based on the input
        # type (where prm \in self.inputs)
        def paramSingle(prm):
            if self.safeGet(prm["id"], "value-choices"):
                return rnd.choice(self.safeGet(prm["id"], "value-choices"))
            if prm["type"] == "String":
                return randStr(prm["id"])
            if prm["type"] == "Number":
                return randNum(prm)
            # Since a flag can't be False, it's either there or not,
            # there's no point in setting it to False.
            if prm["type"] == "Flag":
                return True
            if prm["type"] == "File":
                return randFile(prm["id"])

        # For this function, given prm (a parameter description),
        # a parameter value is generated
        # If prm is a list, a sequence of outputs is generated;
        # otherwise, a single value is returned
        def makeParam(prm):
            mn = self.safeGet(prm["id"], "min-list-entries") or 2
            mx = self.safeGet(prm["id"], "max-list-entries") or nl
            isList = self.safeGet(prm["id"], "list") or False
            return (
                [paramSingle(prm) for _ in range(rnd.randint(mn, mx))]
                if isList
                else paramSingle(prm)
            )

        # Returns a list of the ids of parameters that
        # disable the input parameter
        def disablersOf(inParam):
            return [
                disabler[0]
                for disabler in [
                    disabler
                    for disabler in [
                        (
                            prm["id"],
                            self.safeGet(prm["id"], "disables-inputs") or [],
                        )
                        for prm in self.inputs
                    ]
                    if inParam["id"] in disabler[1]
                ]
            ]

        # Returns the list of mutually requiring parameters of the target
        def mutReqs(targetParam):
            return [
                self.byId(mutualReq[0])
                for mutualReq in [
                    possibleMutReq
                    for possibleMutReq in [
                        (reqOfTarg, self.reqsOf(reqOfTarg))
                        for reqOfTarg in self.reqsOf(targetParam["id"])
                    ]
                    if targetParam["id"] in possibleMutReq[1]
                ]
            ]

        def getAllBranchedReqs(targ, reqs):
            for r in [r for r in self.reqsOf(targ["id"]) if r not in reqs]:
                if r in [g["id"] for g in self.groups]:
                    for gReq in self.reqsOf(r):
                        if gReq not in reqs:
                            reqs.append(gReq)
                            getAllBranchedReqs(self.byId(gReq), reqs)
                elif r not in reqs:
                    reqs.append(r)
                    getAllBranchedReqs(self.byId(r), reqs)
            return reqs

        # Returns whether targ (an input parameter) has a
        # value or is allowed to have one
        def isOrCanBeFilled(targ):
            # If it is already filled in, report so
            if targ["id"] in list(self.in_dict.keys()):
                return True
            # If a disabler or a disabled target is already active,
            # it cannot be filled
            for d in disablersOf(targ) + (
                self.safeGet(targ["id"], "disables-inputs") or []
            ):
                if d in list(self.in_dict.keys()):
                    return False
            # If it is in a mutex group with one target already chosen,
            # it cannot be filled
            for r in self.reqsOf(targ["id"]):
                if r in [g["id"] for g in self.groups]:
                    grpCanBeFilled = False
                    for gReq in self.reqsOf(r):
                        if isOrCanBeFilled(self.byId(gReq)):
                            # if one of members can be added, skip the rest
                            grpCanBeFilled = True
                            continue
                    if not grpCanBeFilled:
                        return False
                elif r not in self.in_dict:  # If a requirement is not present
                    # and it is not mutually required
                    if targ["id"] not in getAllBranchedReqs(targ, []):
                        return False
            # Get the group that the target belongs to, if any
            g = self.assocGrp(targ["id"])
            if (g is not None) and self.safeGrpGet(g["id"], "mutually-exclusive"):
                if len([x for x in g["members"] if x in list(self.in_dict.keys())]) > 0:
                    return False
            return True

        # For optional params, don't add if it requires inputs not
        # already chosen. Requires-complete can be satisfied but is
        # too complicated to check without building dependency tree.
        # (Added after requires-inputs: group was implemented)
        def groupMemberCanBeAdded(targ):
            # if target is a group member
            if targ["id"] in [m for g in self.groups for m in g["members"]]:
                for req in self.reqsOf(targ["id"]):
                    # If targ requires group input and one of required members
                    # has been chosen
                    if (
                        req in [g["id"] for g in self.groups]
                        and len(set(self.reqsOf(req)).intersection(set(self.in_dict)))
                        == 1
                    ):
                        continue
                    elif req not in self.in_dict:
                        return False
                return True
            elif set(self.reqsOf(targ["id"])).issubset(set(self.in_dict)):
                return True
            return False

        # Handle the mutual requirement case by breadth first search in
        # the graph of mutual requirements. Essentially a graph is built,
        # starting from targ (the target input parameter), where nodes
        # are input parameters and edges are a mutual requirement
        # relation between nodes. BFS is used to check every node, to see
        # if can be (or has been) given a value. If all the nodes are
        # permitted to be given values, then they are all added at once;
        # if even one of them cannot be given a value (e.g. it has an
        # active disabler) then none of the graph members can be added
        # and so we just return false. # Input: an input parameter from
        # which to start building the graph Output: Returns False if at
        # least one of the mutual requirements cannot be met Returns a
        # list of params to fill if all of them can be met (or [targ.id]
        # if it has no mutReqs)
        def checkMutualRequirements(targ):
            checked, toCheck = [], [targ]
            while len(toCheck) > 0:
                current = toCheck.pop()
                checked.append(current)
                if not isOrCanBeFilled(current):
                    return False
                for mutreq in mutReqs(current):
                    if not mutreq["id"] in [c["id"] for c in checked] and mutreq[
                        "id"
                    ] not in [g["id"] for g in self.groups]:
                        toCheck.append(mutreq)
                for greq in [
                    g
                    for g in self.reqsOf(current["id"])
                    if g in [grp["id"] for grp in self.groups]
                    and self.safeGrpGet(g, "mutually-exclusive")
                ]:
                    # Check if one of the members is already added
                    # and don't add random member
                    memberFilled = False
                    for m in self.safeGrpGet(greq, "members"):
                        if m in self.in_dict:
                            memberFilled = True
                    if memberFilled:
                        continue
                    # Add random member if current requires mutex group
                    rndMember = rnd.choice(self.safeGrpGet(greq, "members"))
                    toCheck.append({i["id"]: i for i in self.inputs}[rndMember])
            return checked

        # Start actual dictionary filling part
        # Clear the dictionary
        self.in_dict = (
            self.in_dict
            if hasattr(self, "in_dict") and self.in_dict is not None
            else {}
        )
        for params in [r for r in self.inputs if not r.get("optional")]:
            self.in_dict[params["id"]] = makeParam(params)

        # Fill in a random choice for each one-is-required group
        for grp in [
            g for g in self.groups if self.safeGrpGet(g["id"], "one-is-required")
        ]:
            # Loop to choose an allowed value,
            # in case a previous choice disabled that one
            while True:
                # Pick a random parameter
                mbrId = rnd.choice(
                    [mbr for mbr in grp["members"] if self.byId(mbr)["type"] != "Flag"]
                )
                choice = self.byId(mbrId)
                # see if it and its mutual requirements can be filled
                res = checkMutualRequirements(choice)
                if res is False:
                    # Try again if the chosen group member is not permissible
                    continue
                for r in res:
                    self.in_dict[r["id"]] = makeParam(r)
                break  # If we were allowed to add a parameter, we can stop

        if self.requireComplete:
            # Fill in all possible optional inputs
            opts = [
                p
                for p in self.inputs
                if self.safeGet(p["id"], "optional") in [None, True]
            ]
            # Loop a random number of times, each time
            #  attempting to fill a random parameter
            for option in opts:
                # If it is already filled in, continue
                if option["id"] in list(self.in_dict.keys()):
                    continue
                # If it is a prohibited option, continue
                # (isFilled case handled above)
                if not isOrCanBeFilled(option):
                    continue
                # Implementation without building dependency tree
                # (should redo localexec)
                if not groupMemberCanBeAdded(option):
                    continue
                # Now we handle the mutual requirements case. This is a little
                # more complex because a mutual requirement
                # of targ can have its own mutual requirements, ad nauseam.
                # We need to look at all of them recursively and either
                # fill all of them in (i.e. at once) or none of them
                # (e.g. if one of them is disabled by some other param).
                result = checkMutualRequirements(option)
                # Leave if the mutreqs cannot be satisfied
                if result is False:
                    continue
                # Fill in the target(s) otherwise
                for r in result:
                    self.in_dict[r["id"]] = makeParam(r)
                    # Check for mutex between in_dict and last in param
                    for group, mbs in [
                        (x, x["members"])
                        for x in self.groups
                        if x.get("mutually-exclusive")
                    ]:
                        if (
                            len(set.intersection(set(mbs), set(self.in_dict.keys())))
                            > 1
                        ):
                            # Delete last in param
                            del self.in_dict[r["id"]]

    # Function to generate random parameter values
    # This fills the in_dict with random values, validates the input,
    # and generates the appropriate command line
    def generateRandomParams(self, generateCmdLineFromInDict=False):
        """
        The generateRandomParams method fills the in_dict field
        with randomly generated values following the schema.
        It then generates command line strings based on these values
        """

        self.cmd_line = []
        # Set in_dict with random values
        self._randomFillInDict()
        # Look at generated input, if debugging
        if self.debug:
            print_info("Input: " + str(self.in_dict))
        # Check results (as much as possible)
        try:
            args = [self.desc_path, "-i", json.dumps(self.in_dict)]
            if self.sandbox:
                args.append("--sandbox")
            boutiques.invocation(*args)
        # If an error occurs, print out the problems already
        # encountered before blowing up
        except Exception as e:  # Avoid BaseExceptions like SystemExit
            sys.stderr.write(
                "An error occurred in validation\n" "Previously saved issues\n"
            )
            for err in self.errs:
                sys.stderr.write("\t" + str(err) + "\n")
            raise e  # Pass on (throw) the caught exception
        if generateCmdLineFromInDict:
            # Add new command line
            self.cmd_line.append(self._generateCmdLineFromInDict())

    # Read in parameter input file or string
    def readInput(self, infile):
        """
        The readInput method sets the in_dict field of the executor
        object, based on a fixed input.
        It then generates a command line based on the input.

        infile: either the inputs in a file or
        the command-line string (from -s).
        stringInput: a boolean as to whether the method has
        been given a string or a file.
        """

        # Quick check that the descriptor has already been read in
        assert self.desc_dict is not None
        self.in_dict = loadJson(infile)

        # Input dictionary
        if self.debug:
            print_info("Input: " + str(self.in_dict))

        # Generate public invocation
        if not self.skipDataCollect:
            self.public_in = self._generatePublicInvocation()

        # Add default values for required parameters,
        # if no value has been given
        addDefaultValues(self.desc_dict, self.in_dict)
        # Check results (as much as possible)
        try:
            args = [self.desc_path, "-i", json.dumps(self.in_dict)]
            if self.sandbox:
                args.append("--sandbox")
            boutiques.invocation(*args)
        except Exception:  # Avoid catching BaseExceptions like SystemExit
            sys.stderr.write(
                "An error occurred in validation\n" "Previously saved issues\n"
            )
            for err in self.errs:
                # Write any errors we found
                sys.stderr.write("\t" + str(err) + "\n")
            raise  # Raise the exception that caused failure
        # Build and save output command line (as a single-entry list)
        self.cmd_line = [self._generateCmdLineFromInDict()]

    # Private method to replace the keys in template by input and output
    # values. Input and output values are looked up in self.in_dict and
    # self.out_dict
    # * if use_flags is true, keys will be replaced by:
    #      * flag+flag-separator+value if flag is not None
    #      * value otherwise
    # * if unfound_keys is "remove", unfound keys will be replaced by ""
    # * if unfound_keys is "clear" then the template is cleared if it has
    #     unfound keys (useful for configuration files)
    # * before being substituted, the values will be:
    #     * stripped from all the strings in stripped_extensions
    #     * escaped for special characters
    def _replaceKeysInTemplate(
        self,
        template,
        use_flags=False,
        unfound_keys="remove",
        stripped_extensions=[],
        is_output=False,
        escape_special_chars=True,
    ):
        def escape_string(s):
            try:
                from shlex import quote
            except ImportError as e:
                from shlex import quote
            return quote(s)

        # Concatenate input and output dictionaries
        in_out_dict = dict(self.in_dict)
        in_out_dict.update(self.out_dict)
        # Go through all the keys
        for param_id in [x["id"] for x in self.inputs + self.outputs]:
            escape = (
                escape_special_chars
                and (
                    self.safeGet(param_id, "type") == "String"
                    or self.safeGet(param_id, "type") == "File"
                )
                or param_id in self.out_dict.keys()
            )
            clk = self.safeGet(param_id, "value-key")
            if clk is None:
                continue
            if param_id in list(in_out_dict.keys()):  # param has a value
                val = in_out_dict[param_id]
                if type(val) is list:
                    list_sep = self.safeGet(param_id, "list-separator")
                    if list_sep is None:
                        list_sep = " "
                    escaped_val = []
                    for x in val:
                        escaped_val.append(escape_string(str(x)) if escape else str(x))
                    val = list_sep.join(escaped_val)
                elif escape:
                    val = escape_string(val)
                # Add flags and separator if necessary
                flag = self.safeGet(param_id, "command-line-flag")
                if use_flags and flag is not None:
                    sep = self.safeGet(param_id, "command-line-flag-separator")
                    if sep is None:
                        sep = " "
                    # special case for flag-type inputs
                    if self.safeGet(param_id, "type") == "Flag":
                        val = "" if val is False else flag
                    else:
                        val = flag + sep + str(val)
                # Remove file extensions from input value
                if (
                    self.safeGet(param_id, "type") == "File"
                    or self.safeGet(param_id, "type") == "String"
                ):
                    for extension in stripped_extensions:
                        val = val.replace(extension, "")
                    # Remove path if a) a file, b) not the first item in the
                    # template; for output files specifically
                    if (
                        self.safeGet(param_id, "type") == "File"
                        and template.find(clk) > 0
                        and is_output
                    ):
                        val = op.basename(val)
                # Here val can be a number so we need to cast it
                if val is not None and val != "":
                    template = template.replace(clk, str(val))
                else:
                    template = template.replace(" " + clk, str(val))
            else:  # param has no value
                if unfound_keys == "remove":
                    if (" " + clk) in template:
                        template = template.replace(" " + clk, "")
                    else:
                        template = template.replace(clk, "")
                elif unfound_keys == "clear":
                    if clk in template:
                        return ""
        return template

    # Private method to generate output file names.
    # Output file names will be put in self.out_dict.
    def _generateOutputFileNames(self):
        if not hasattr(self, "out_dict"):
            # a dictionary that will contain the output file names
            self.out_dict = {}
        for outputId, isPathTemplate in [
            (x["id"], "path-template" in x) for x in self.outputs
        ]:
            if isPathTemplate:
                if outputId in list(self.out_dict.keys()):
                    outputFileName = self.out_dict[outputId]
                else:
                    outputFileName = self.safeGet(outputId, "path-template")

            # if 'conditional-path-template' in outputItem
            # (key=conditions, value=path)
            # Initialize file name with path template or existing value
            elif not isPathTemplate:
                for templateObj in self.safeGet(outputId, "conditional-path-template"):
                    templateKey = list(templateObj.keys())[0]
                    condition = self._getCondPathTemplateExp(templateKey)
                    # If condition is true, set fileName
                    # Stop checking (if-elif...)
                    if condition == "default":
                        outputFileName = templateObj["default"]
                        break
                    elif eval(condition):
                        outputFileName = templateObj[templateKey]
                        break

            stripped_extensions = self.safeGet(
                outputId, "path-template-stripped-extensions"
            )
            if stripped_extensions is None:
                stripped_extensions = []
            se = stripped_extensions  # Renaming variable to save space
            # We keep the unfound keys because they will be
            # substituted in a second call to the method in case
            # they are output keys
            outputFileName = self._rkit(
                outputFileName,
                use_flags=False,
                unfound_keys="keep",
                stripped_extensions=se,
                is_output=True,
                escape_special_chars=False,
            )

            if self.safeGet(outputId, "uses-absolute-path"):
                outputFileName = os.path.abspath(outputFileName)
            self.out_dict[outputId] = outputFileName

    def _getCondPathTemplateExp(self, templateKey):
        splitExp = conditionalExpFormat(templateKey).split()
        parsedExp = []
        for word in [word.strip() for word in splitExp if len(word) > 0]:
            all_ids = [i["id"] for i in (self.inputs + self.outputs)]
            # Substitute boolean expression key by its value
            in_out_dict = self.in_dict.copy()
            in_out_dict.update(self.out_dict)
            if word in in_out_dict:
                value = f"{in_out_dict[word]}"
                if value.replace(".", "").replace("-", "").isdigit():
                    parsedExp.append(in_out_dict[word])
                else:
                    parsedExp.append(f'"{value}"')
            # Boolean expression key is not chosen (optional input),
            # therefore expression is false
            elif word in all_ids:
                parsedExp = ["False"]
                break
            # Word is an expression char, just append it
            else:
                parsedExp.append(word)
        return " ".join(f"{w}" for w in parsedExp)

    # Private method to write configuration files
    # Configuration files are output files that have a file-template
    def _writeConfigurationFiles(self):
        for outputId in [x["id"] for x in self.outputs]:
            fileTemplate = self.safeGet(outputId, "file-template")
            if fileTemplate is None:
                continue  # this is not a configuration file
            stripped_extensions = self.safeGet(
                outputId, "path-template-stripped-extensions"
            )
            if stripped_extensions is None:
                stripped_extensions = []
            se = stripped_extensions  # Renaming variable to save space
            # We substitute the keys line by line so that we can
            # clear the lines that have keys with no value
            # (undefined optional params)
            newTemplate = []
            for line in fileTemplate:
                newTemplate.append(
                    self._rkit(
                        line,
                        use_flags=False,
                        unfound_keys="clear",
                        stripped_extensions=se,
                        is_output=False,
                        escape_special_chars=True,
                    )
                )
            template = os.linesep.join(newTemplate)
            # Write the configuration file
            fileName = self.out_dict[outputId]
            dirs = os.path.dirname(fileName)
            if dirs and not os.path.exists(dirs):
                os.makedirs(dirs)
            with open(fileName, "w+") as fil:
                fil.write(template)

    # Private method to build the actual command line by substitution,
    # using the input data
    def _generateCmdLineFromInDict(self):
        # Generate output file names
        self._generateOutputFileNames()
        # it is required to call the method twice in case path
        # templates contain output keys
        self._generateOutputFileNames()
        # Write configuration files
        self._writeConfigurationFiles()
        # Get the command line template
        template = self.desc_dict["command-line"]
        # Substitute every given value into the template
        # (incl. flags, flag-seps, ...)
        template = self._rkit(
            template,
            use_flags=True,
            unfound_keys="remove",
            stripped_extensions=[],
            is_output=False,
            escape_special_chars=True,
        )
        # Return substituted command line
        return template

    # Print the command line result
    def printCmdLine(self):
        print("Generated Command" + ("s" if len(self.cmd_line) > 1 else "") + ":")
        for cmd in self.cmd_line:
            print(cmd)

    # Private method to generate summary object of data
    # collection file containing the tool name and descriptor DOI
    def _generateSummary(self, desc):
        summary = {}
        summary["name"] = self.desc_dict["name"]
        summary["descriptor-doi"] = self._findDOI(desc)
        return summary

    # Private method to attempt to find descriptor DOI
    # through various cases
    # userInput may be json string, filename or zenodo id
    def _findDOI(self, userIn):
        doi_prefix = "10.5281/"
        # DOI in Zenodo reference case
        if userIn.startswith(doi_prefix):
            return userIn
        elif userIn.split(".")[0].lower() == "zenodo":
            return doi_prefix + userIn
        # File cases
        if os.path.isfile(userIn):
            # Most recent DOI in file if user is publisher
            # Include check to ensure descriptor is unmodified
            if self.desc_dict.get("doi") is not None:
                # Popping the DOI allows it to match published version.
                # In a match, we'll re-add the DOI
                doi = self.desc_dict.pop("doi")
                if loadJson(doi) == self.desc_dict:
                    self.desc_dict["doi"] = doi
                    return doi
            # DOI in filename if descriptor pulled from Zenodo
            # Include check to ensure descriptor is as published
            elif os.path.basename(userIn).split("-")[0].lower() == "zenodo":
                doi = os.path.basename(userIn).split(".")[0].replace("-", ".")
                if loadJson(doi) == self.desc_dict:
                    return doi_prefix + doi
        # No DOI found, save descriptor to cache and return filename
        return self._saveDescriptorToCache()

    # Private method to generate public invocation object
    # for data collection file
    # absolute paths are stripped to filenames and hashes are generated for
    # each input of type File, all other inputs are recorded as submitted
    def _generatePublicInvocation(self):
        public_in_dict = self.in_dict.copy()
        # Replace file type inputs with object containing
        # input file hash and filename
        for x in self.inputs:
            if x.get("type") == "File":
                id = x.get("id")
                path = public_in_dict.get(id)
                if path is not None:
                    if isinstance(path, list):
                        public_in_dict[id] = [self._buildPublicFile(p) for p in path]
                    else:
                        public_in_dict[id] = self._buildPublicFile(path)

        return public_in_dict

    # Private method to generate public output object for data collection file
    # hashes are generated for each output file.
    def _generatePublicOutput(self, exec_output, out_files_dict, missing_files_dict):
        public_out_dict = {}
        public_out_dict["stdout"] = exec_output.stdout
        public_out_dict["stderr"] = exec_output.stderr
        public_out_dict["exit-code"] = exec_output.exit_code
        public_out_dict["error-message"] = exec_output.error_message
        public_out_dict["shell-command"] = exec_output.shell_command
        public_out_dict["missing-files"] = missing_files_dict

        # Iterate through output files to generate output objects
        # and generate objects with hash of files
        out_files_dict = {
            id: self._buildPublicFile(filename)
            for id, filename in out_files_dict.items()
        }
        public_out_dict["output-files"] = out_files_dict
        return public_out_dict

    # Private method to recursively explore directory and hash all files
    def _buildPublicFile(self, path):
        filename = extractFileName(path)
        # If path is not found, report it
        if not os.path.exists(path):
            return {"file-name": filename, "not_found": True}
        # Directories are expanded recursively
        if os.path.isdir(path):
            contents = os.listdir(path)
            # Recursive call to expand directory
            files = [self._buildPublicFile(os.path.join(path, x)) for x in contents]
            return {"file-name": filename, "files": files}
        # Files are hashed
        else:
            md5sum = computeMD5(path)
            return {"file-name": filename, "md5sum": md5sum}

    # Private method to publish data collection objects to file
    # summary, publicInput an publicOutput are combined and
    # written to a file in .cache
    def _saveDataCaptureToCache(self):
        date_time = datetime.datetime.now().strftime("%Y-%m-%d_%Hh%Mm%Ss%fms")
        tool_name = self.summary["name"].replace(" ", "-")
        self.summary["date-time"] = date_time
        # Combine three modules plus provenance in master dictionary
        data_dict = {
            "summary": self.summary,
            "public-invocation": self.public_in,
            "public-output": self.public_out,
            "additional-information": self.provenance,
        }
        # Convert dictionary to Json string
        content = json.dumps(data_dict, indent=4)
        # Write collected data to file
        data_cache_dir = getDataCacheDir()
        filename = f"{tool_name}_{date_time}.json"
        file_path = os.path.join(data_cache_dir, filename)
        file = open(file_path, "w+")
        file.write(content)
        file.close()
        if self.debug:
            print_info(f"Data capture from execution saved to cache as {filename}")

    # Local function handles case where descriptor is not published
    # Checks if descriptor already saved to cache, if not then saves
    # copy for future publication
    def _saveDescriptorToCache(self):
        tool_name = self.desc_dict.get("name").replace(" ", "-")
        data_cache_dir = getDataCacheDir()
        data_cache_files = os.listdir(data_cache_dir)
        # Filter for descriptors in cache with the same tool name to check
        # if descriptor already in cache
        matching_files = [
            x
            for x in data_cache_files
            if len(x.split("_")) > 1 and x.split("_")[1] is tool_name.replace(" ", "-")
        ]
        match = None
        for fl in matching_files:
            path = os.path.join(data_cache_dir, fl)
            file_dict = loadJson(path)
            if file_dict == self.desc_dict:
                match = fl
                break
        if match:
            if self.debug:
                print_info(
                    f"Unpublished descriptor match found in data cache as {match}"
                )
            return match
        # Write descriptor to data cache and save return filename
        content = json.dumps(self.desc_dict, indent=4)
        date_time = datetime.datetime.now().strftime("%Y-%m-%d_%Hh%Mm%Ss%fms")
        filename = f"descriptor_{tool_name}_{date_time}.json"
        path = os.path.join(data_cache_dir, filename)
        file = open(path, "w+")
        file.write(content)
        file.close()
        if self.debug:
            print_info(
                "Descriptor from execution saved to cache for future "
                "publishing as {}".format(filename)
            )
        return filename


# Adds default values to input dictionary
# for parameters whose values were not given
def addDefaultValues(desc_dict, in_dict):
    inputs = desc_dict["inputs"]
    for in_param in [s for s in inputs if s.get("default-value") is not None]:
        if in_dict.get(in_param["id"]) is None:
            in_dict[in_param["id"]] = in_param.get("default-value")
    return in_dict


# Hashes files with MD5,
# capable of handling large data files
def computeMD5(filepath):
    hash_md5 = hashlib.md5()
    with open(filepath, "rb") as fhandle:
        for chunk in iter(lambda: fhandle.read(4096), b""):
            hash_md5.update(chunk)
    return hash_md5.hexdigest()



#!/usr/bin/env python

import os
import os.path as op
import subprocess
from argparse import ArgumentParser

import pytest
import simplejson as json

import boutiques.creator as bc
from boutiques import __file__ as bfile, bosh
from boutiques.creator import CreatorError
from boutiques.tests.BaseTest import BaseTest
from boutiques.util.utils import loadJson


class TestCreator(BaseTest):
    @pytest.fixture(autouse=True)
    def set_test_dir(self):
        self.setup("creator")

    def test_success_template(self):
        fil = "./test_temp/creator_output.json"
        descriptor = bosh(["create", fil])
        self.assertIsNone(bosh(["validate", fil]))

    def test_success_docker(self):
        fil = "./test_temp/creator_output.json"
        descriptor = bosh(["create", "-d", "mysql:latest", fil])
        self.assertIsNone(bosh(["validate", fil]))

    def test_success_docker_sing_import(self):
        fil = "./test_temp/creator_output.json"
        descriptor = bosh(["create", "-d", "mysql:latest", "-u", fil])
        self.assertIsNone(bosh(["validate", fil]))

    @pytest.mark.skipif(
        subprocess.Popen("type docker", shell=True).wait(),
        reason="Docker not installed",
    )
    def test_fail_image_(self):
        fil = "./test_temp/creator_output.json"
        self.assertRaises(
            CreatorError, bosh, ["create", "-d", "ihopethisdoesntexists", fil]
        )

    def test_not_an_argparser(self):
        self.assertRaises(CreatorError, bc.CreateDescriptor, "notaparser")

    def test_success_argparser(self):
        parser = ArgumentParser(description="my tool description")
        parser.add_argument("myarg1", action="store", help="my help 1", type=list)
        parser.add_argument("myarg2", action="store", help="my help 2", type=int)
        parser.add_argument("--myarg3", "-m", action="store", help="my help 3")
        subparser = parser.add_subparsers(
            help="the choices you will make", dest="mysubparser"
        )
        sb1 = subparser.add_parser("option1", help="the first value")
        sb1.add_argument("suboption1", help="the first sub option option")
        sb1.add_argument(
            "suboption2",
            help="the first sub option option",
            choices=["hello", "goodbye"],
            default="hello",
        )

        sb1 = subparser.add_parser("option2", help="the second value")
        sb1.add_argument("suboption1", help="the first sub option option")
        sb1.add_argument("suboption3", help="the second sub option option")
        sb1.add_argument("--suboptionflag1", "-s", help="the bool opt flag")
        sb1.add_argument(
            "--suboptionflag2",
            "-d",
            action="store_true",
            help="the second sub option flag",
        )
        sb1.add_argument("--suboptlistflag", "-l", nargs="+", help="listy flag")

        creatorObj = bc.CreateDescriptor(
            parser,
            execname="/path/to/myscript.py",
            verbose=True,
            tags={"purpose": "testing-creator", "foo": "bar"},
        )

        fil = "./test_temp/test-created-argparse-descriptor.json"
        creatorObj.save(fil)

        invof = "./test_temp/test-created-argparse-inputs.json"
        args = parser.parse_args(
            [
                ["val1", "val2"],
                "2",
                "option2",
                "subval1",
                "subval3",
                "--suboptionflag1",
                "t1",
                "--suboptionflag2",
            ]
        )
        invo = creatorObj.createInvocation(args)
        with open(invof, "w") as fhandle:
            fhandle.write(json.dumps(invo, indent=4))

        self.assertIsNone(bosh(["validate", fil]))
        self.assertIsNone(bosh(["invocation", fil, "-i", invof]))

    def test_success_template_camel_case(self):
        template = op.join(op.split(self.tests_dir)[0], "templates", "basic.json")
        fil = "./test_temp/creator_output.json"
        bosh(["create", fil, "--camel-case"])
        self.assertIsNone(bosh(["validate", fil]))

        desc = loadJson(fil)
        template = loadJson(template)

        # Check "_" in all instances of input indices (inputs + groups)
        for inp, camelInp in zip(template["inputs"], desc["inputs"]):
            self.assertTrue("_" in inp["id"])
            self.assertFalse("_" in camelInp["id"])

        for mbrs, camelMbrs in [
            (grp["members"], camelGrp["members"])
            for grp, camelGrp in zip(template["groups"], desc["groups"])
        ]:
            self.assertTrue(all([("_" in mbr) for mbr in mbrs]))
            self.assertFalse(all([("_" in camelMbr) for camelMbr in camelMbrs]))

    def test_create_cl_template_from_descriptor(self):
        cl_template = self.get_file_path("expected_cl_template_create.json")
        output = self.get_file_path("out_desc.json")
        expected_cml = loadJson(cl_template)["command-line"]
        expected_inputs = loadJson(cl_template)["inputs"]

        create_args = ["create", output, "--cl-template", cl_template]
        bosh(create_args)
        result_cml = loadJson(output)["command-line"]
        result_inputs = loadJson(output)["inputs"]

        if op.exists(output):
            os.remove(output)

        self.assertEqual(expected_cml, result_cml)
        self.assertEqual(expected_inputs, result_inputs)

    def test_create_cl_template_from_string(self):
        cl_template = self.get_file_path("expected_cl_template_create.json")
        output = self.get_file_path("out_desc.json")
        expected_cml = "echo [PARAM1] [PARAM2] [FLAG1] > [OUTPUT1]"
        expected_inputs = loadJson(cl_template)["inputs"]

        create_args = ["create", output, "--cl-template", expected_cml]
        bosh(create_args)
        result_cml = loadJson(output)["command-line"]
        result_inputs = loadJson(output)["inputs"]

        if op.exists(output):
            os.remove(output)

        self.assertEqual(expected_cml, result_cml)
        self.assertEqual(expected_inputs, result_inputs)



#!/usr/bin/env python

import numbers
import sys
from collections import OrderedDict
from operator import itemgetter
from urllib.parse import quote

from boutiques.logger import print_info, raise_error
from boutiques.publisher import ZenodoError
from boutiques.zenodoHelper import ZenodoHelper


class Searcher:

    def __init__(
        self,
        query,
        verbose=False,
        sandbox=False,
        max_results=None,
        no_trunc=False,
        exact_match=False,
    ):
        if query is not None:
            self.query = query
            self.query_line = ""
            if not exact_match:
                terms = self.query.replace("/", ".").split(" ")
                for t in terms:
                    uncased_term = [
                        f"[{ch.upper() + ch.lower() if ch.isalpha() else ch}]"
                        for ch in t
                    ]
                    uncased_term = quote("".join(uncased_term))
                    self.query_line += f" AND (/.*{uncased_term}.*/)"
            else:
                self.query_line = f" AND (/{self.query.replace('/', '.')}/)"
        else:
            self.query = ""
            self.query_line = ""

        if verbose:
            print_info("Using Query Line: " + self.query_line)

        self.verbose = verbose
        self.sandbox = sandbox
        self.no_trunc = no_trunc
        self.max_results = max_results
        self.zenodo_helper = ZenodoHelper(sandbox=self.sandbox, verbose=self.verbose)

        # Display top 10 results by default
        if max_results is None:
            self.max_results = 10

        # Zenodo will error if asked for more than 9999 results
        if self.max_results > 9999:
            self.max_results = 9999

        # Set Zenodo endpoint
        self.zenodo_endpoint = (
            "https://sandbox.zenodo.org" if self.sandbox else "https://zenodo.org"
        )
        if self.verbose:
            print_info(f"Using Zenodo endpoint {self.zenodo_endpoint}")

    def search(self):
        results = self.zenodo_helper.zenodo_search(self.query, self.query_line)
        total_results = results.json()["hits"]["total"]
        total_deprecated = len(
            [
                h["metadata"]["keywords"]
                for h in results.json()["hits"]["hits"]
                if "metadata" in h
                and "keywords" in h["metadata"]
                and "deprecated" in h["metadata"]["keywords"]
            ]
        )
        results_list = (
            self.create_results_list_verbose(results.json())
            if self.verbose
            else self.create_results_list(results.json())
        )
        num_results = len(results_list)
        print_info(
            "Showing %d of %d result(s)%s"
            % (
                (num_results if num_results < self.max_results else self.max_results),
                (total_results if self.verbose else total_results - total_deprecated),
                (
                    "."
                    if self.verbose
                    else ", excluding %d deprecated result(s)." % total_deprecated
                ),
            )
        )
        return results_list

    def create_results_list(self, results):
        results_list = []
        for hit in results["hits"]["hits"]:
            (id, title, description, downloads) = self.parse_basic_info(hit)
            # skip hit if result is deprecated
            keyword_data = self.get_keyword_data(hit["metadata"]["keywords"])
            if "deprecated" in keyword_data["other"]:
                continue
            result_dict = OrderedDict(
                [
                    ("ID", id),
                    ("TITLE", title),
                    ("DESCRIPTION", description),
                    ("DOWNLOADS", downloads),
                ]
            )
            if not self.no_trunc:
                result_dict = self.truncate(result_dict, 40)
            results_list.append(result_dict)
        results_list = sorted(results_list, key=itemgetter("DOWNLOADS"), reverse=True)
        # Truncate the list according to the desired maximum number of results
        return results_list[: self.max_results]

    def create_results_list_verbose(self, results):
        results_list = []
        for hit in results["hits"]["hits"]:
            (id, title, description, downloads) = self.parse_basic_info(hit)
            author = hit["metadata"]["creators"][0]["name"]
            version = hit["metadata"].get("version", "unknown")
            publication_date = hit["metadata"]["publication_date"]
            doi = hit["doi"]
            keyword_data = self.get_keyword_data(hit["metadata"]["keywords"])
            schema_version = keyword_data["schema-version"]
            container = keyword_data["container-type"]
            other_tags = ",".join(keyword_data["other"])
            result_dict = OrderedDict(
                [
                    ("ID", id),
                    ("TITLE", title),
                    ("DESCRIPTION", description),
                    ("PUBLICATION DATE", publication_date),
                    ("DEPRECATED", "deprecated" in keyword_data["other"]),
                    ("DOWNLOADS", downloads),
                    ("AUTHOR", author),
                    ("VERSION", version),
                    ("DOI", doi),
                    ("SCHEMA VERSION", schema_version),
                    ("CONTAINER", container),
                    ("TAGS", other_tags),
                ]
            )
            if sys.stdout.encoding.lower != "UTF-8":
                for k, v in list(result_dict.items()):
                    if isinstance(v, str):
                        result_dict[k] = v.encode("ascii", "xmlcharrefreplace").decode()
            if not self.no_trunc:
                result_dict = self.truncate(result_dict, 40)
            results_list.append(result_dict)
        results_list = sorted(results_list, key=itemgetter("DOWNLOADS"), reverse=True)
        # Truncate the list according to the desired maximum number of results
        return results_list[: self.max_results]

    def parse_basic_info(self, hit):
        id = "zenodo." + str(hit["id"])
        title = hit["metadata"]["title"]
        description = hit["metadata"]["description"]
        downloads = 0
        if "version_downloads" in hit["stats"]:
            downloads = hit["stats"]["version_downloads"]
        return (id, title, description, downloads)

    # truncates every value of a dictionary whose length is
    # greater than max_length
    def truncate(self, d, max_length):
        for k, v in list(d.items()):
            if isinstance(v, numbers.Number):
                v = str(v)
            if len(v) > max_length:
                d[k] = v[:max_length] + "..."
        return d

    def get_keyword_data(self, keywords):
        keyword_data = {"container-type": "None", "other": []}
        for keyword in keywords:
            if keyword.split(":")[0] == "schema-version":
                keyword_data["schema-version"] = keyword.split(":")[1]
            elif keyword.lower() == "docker" or keyword.lower() == "singularity":
                keyword_data["container-type"] = keyword
            else:
                keyword_data["other"].append(keyword)
        return keyword_data



#!/usr/bin/env python

import keyword
import os.path as op
import re
from argparse import ArgumentParser

import simplejson as json
from jsonschema import ValidationError, validate

from boutiques import __file__ as bfile
from boutiques.logger import print_info, raise_error
from boutiques.util.utils import (
    conditionalExpFormat,
    customSortDescriptorByKey,
    loadJson,
)


# An exception class specific to descriptors
class DescriptorValidationError(ValidationError):
    pass


# Main validation module
def validate_descriptor(descriptor, **kwargs):
    """
    Validates the Boutiques descriptor against the schema.
    """
    path, fil = op.split(bfile)
    schema_file = op.join(path, "schema", "descriptor.schema.json")

    # Load schema
    with open(schema_file) as fhandle:
        schema = json.load(fhandle)

    # Load input types according to the schema
    schema_types = schema["properties"]["inputs"]["items"]["properties"]["type"]["enum"]
    allowed_keywords = ["and", "or", "false", "true"]
    allowed_comparators = ["==", "!=", "<", ">", "<=", ">="]

    # Validate basic JSON schema compliance for descriptor
    # Note: if it fails basic schema compliance we don"t do more checks
    try:
        validate(descriptor, schema)
    except ValidationError as e:
        raise_error(DescriptorValidationError, (str(e)))

    # Helper get functions
    def safeGet(desc, sec, targ):
        if desc.get(sec):
            return [
                item.get(targ) for item in desc[sec] if list(item.keys()).count(targ)
            ]
        return []

    def inputGet(s):
        return safeGet(descriptor, "inputs", s)

    def outputGet(s):
        return safeGet(descriptor, "output-files", s)

    def groupGet(s):
        return safeGet(descriptor, "groups", s)

    def inById(i):
        if i in inputGet("id"):
            return descriptor["inputs"][inputGet("id").index(i)]
        return {}

    def isValidConditionalExp(exp):
        # Return the type of a conditional expression's substring
        def getSubstringType(s):
            s = s.strip()
            if s in schema_types:
                # Can't realistically distinguish File from String
                return s if s != "File" else "String"
            elif re.search(r"^[0-9]*\.?[0-9]+$", s):
                return "Number"
            elif re.search(r"^(True|False|false|true)$", s):
                return "Flag"
            else:
                return "String"

        # Recursively check boolean expression by replacing variables with
        # their expected value type [Number, String, File, Flag]
        brackets, startIdx, endIdx = 0, 0, 0
        rebuiltExp = ""
        for idx, c in enumerate(exp):
            if c == "(":
                brackets += 1
                startIdx = idx + 1
            elif c == ")":
                brackets -= 1
                if brackets == 0:
                    endIdx = idx
                    isValidSubExp = isValidConditionalExp(exp[startIdx:endIdx])
                    # Immediately return false if sub expression is not valid
                    if not isValidSubExp:
                        return False
                    else:
                        rebuiltExp += f"{isValidSubExp}"
            elif brackets == 0:
                rebuiltExp += f"{c}"
        rebuiltExp = rebuiltExp.strip()

        # If there are no more parentheses, check if sub-expression is valid
        # rebuiltExp should only be two values separated by operator: x >= y
        # or values separated by 'AND'/'OR': x and y and z
        if "(" not in rebuiltExp and ")" not in rebuiltExp:
            expElements = rebuiltExp.split()
            for idx, e in enumerate(expElements):
                e = e.strip()
                # Compare types of elements neighbouring allowed_comparators
                if e in allowed_comparators and getSubstringType(
                    expElements[idx - 1]
                ) != getSubstringType(expElements[idx + 1]):
                    return False
                # Check if keyword element is part of allowed keywords
                if keyword.iskeyword(e) and e.lower() not in allowed_keywords:
                    return False
                # Check elements neighbouring and/or keywords are valid words
                if e.lower() in ["and", "or"] and (
                    expElements[idx - 1] not in schema_types + ["True", "False"]
                    or expElements[idx + 1] not in schema_types + ["True", "False"]
                ):
                    return False
        return True

    # Begin looking at Boutiques-specific failures
    errors = []

    clkeys = inputGet("value-key") + outputGet("value-key")
    flattenedTemplates = [y for x in outputGet("file-template") for y in x]
    configFileTemplates = flattenedTemplates + outputGet("path-template")

    cmdline = descriptor["command-line"]

    # Verify that all command-line key appear in the command-line, in
    # a file template or in an environment variable value
    msg_template = (
        '   KeyError: "{0}" not in command-line or file template'
        " or environment variables"
    )
    envValues = ""
    if descriptor.get("environment-variables"):
        for env in descriptor.get("environment-variables"):
            envValues += "||" + env["value"]
    errors += [
        msg_template.format(k)
        for k in clkeys
        if ((cmdline + ".".join(configFileTemplates) + envValues).count(k)) < 1
    ]

    # Verify that no key contains another key
    msg_template = '   KeyError: "{0}" contains "{1}"'
    errors += [
        msg_template.format(key, clkeys[jdx])
        for idx, key in enumerate(clkeys)
        for jdx in range(0, len(clkeys))
        if clkeys[jdx] in key and key != clkeys[jdx]
    ]

    # Verify that all Ids are unique
    inIds, outIds = inputGet("id"), outputGet("id")
    grpIds = groupGet("id") if "groups" in descriptor.keys() else []
    allIds = inIds + outIds + grpIds
    msg_template = '    IdError: "{0}" is non-unique'
    for idx, s1 in enumerate(allIds):
        for jdx, s2 in enumerate(allIds):
            if s1 == s2 and idx < jdx:
                errors += [msg_template.format(s1)]
            else:
                errors += []

    # Verify that identical keys only exist if they are both in mutex groups
    msg_template = ' MutExError: "{0}" belongs to 2+ non exclusive IDs'
    for idx, key in enumerate(clkeys):
        for jdx in range(idx + 1, len(clkeys)):
            if clkeys[jdx] == key:
                mids = [
                    inById(mid)["id"]
                    for mid in inIds
                    if inById(mid)["value-key"] == key
                ]
                for idx, grp in enumerate(descriptor.get("groups")):
                    mutex = grp.get("mutually-exclusive")
                    if set(grp["members"]) == set(mids) and not mutex:
                        errors += [msg_template.format(key)]

    # Verify that output files have unique path-templates
    msg_template = 'OutputError: "{0}" and "{1}" have the same ' "path-template"
    for ix, o1 in zip(outputGet("id"), outputGet("path-template")):
        for jx, o2 in zip(outputGet("id"), outputGet("path-template")):
            if o1 == o2 and jx != ix:
                errors += [msg_template.format(ix, jx)]
            else:
                errors += []

    if "output-files" in descriptor:
        # Verify output file with non-optional conditional file template
        # contains a default path
        msg_template = (
            'OutputError: "{0}". Non-optional output-file with '
            "conditional-path-template must contain "
            '"default" path-template.'
        )

        cond_outfiles_keys = []
        for outF in [
            o
            for o in descriptor["output-files"]
            if "conditional-path-template" in o and not o["optional"]
        ]:
            out_keys = [
                list(obj.keys())[0] for obj in outF["conditional-path-template"]
            ]
            cond_outfiles_keys.extend(out_keys)
            if "default" not in out_keys:
                errors += [msg_template.format(outF["id"])]

            # Verify output keys contain only one default condition
            if out_keys.count("default") > 1:
                errors += [
                    'OutputError: "{}". Only one "default" '
                    "condition is permitted in a "
                    "conditional-path-template.".format(outF["id"])
                ]

        # Verify output key contains variables that correspond to input IDs
        # or is 'default'
        msg_template = (
            'OutputError: "{0}" contains non-python keyword and ' 'non-ID string: "{1}"'
        )
        for templateKey in cond_outfiles_keys:
            splitExp = conditionalExpFormat(templateKey).split()
            if splitExp[0] == "default" and len(splitExp) == 1:
                continue
            for s in [
                s
                for s in splitExp
                if not keyword.iskeyword(s)
                and s.isalnum()
                and not s.isdigit()
                and s not in [i["id"] for i in descriptor["inputs"]]
                and s not in [i["id"] for i in descriptor["output-files"]]
            ]:
                errors += [msg_template.format(outF["id"], s)]

        # Verify variable is being evaluated against a value of the same type
        msg_template = (
            'OutputError: Conditional output "{0}" contains '
            "invalid conditional expression. Verify arguments' "
            "type and allowed keywords in expression."
        )
        for templateKey in cond_outfiles_keys:
            splitExp = conditionalExpFormat(templateKey).split()
            if splitExp[0] == "default" and len(splitExp) == 1:
                continue
            # Replace variable by it's type according to the descriptor schema
            for s in [
                s
                for s in enumerate(splitExp)
                if not keyword.iskeyword(s[1]) and s[1].isalnum() and not s[1].isdigit()
            ]:
                if s[1] in [i["id"] for i in descriptor["inputs"]]:
                    splitExp[s[0]] = inById(s[1])["type"]
            # Check if the conditional expression is valid
            if not isValidConditionalExp(" ".join(splitExp)):
                errors += [msg_template.format(templateKey)]

    # Verify inputs
    for inp in descriptor["inputs"]:

        # Add optional property in case it's not
        # there (default to false as in JSON)
        if "optional" not in inp.keys():
            inp["optional"] = False

        # Verify flag-type inputs (have flags, not required, cannot be lists)
        if inp["type"] == "Flag":
            msg_template = ' InputError: "{0}" must have a command-line flag'
            if "command-line-flag" not in inp.keys():
                errors += [msg_template.format(inp["id"])]
            else:
                errors += []

            msg_template = (
                ' InputError: "{0}" is of type Flag,' " it has to be optional"
            )
            if inp["optional"] is False:
                errors += [msg_template.format(inp["id"])]
            else:
                errors += []

        # Verify number-type inputs min/max are sensible
        elif inp["type"] == "Number":
            msg_template = (
                ' InputError: "{0}" cannot have greater' " min ({1}) than max ({2})"
            )
            minn = inp["minimum"] if "minimum" in inp.keys() else -float("Inf")
            maxx = inp["maximum"] if "maximum" in inp.keys() else float("Inf")
            if minn > maxx:
                errors += [msg_template.format(inp["id"], minn, maxx)]
            else:
                errors += []

        # Verify enum-type inputs (at least 1 option, default in set)
        elif "value-choices" in inp.keys():
            msg_template = ' InputError: "{0}" must have at least' " one value choice"
            if len(inp["value-choices"]) < 1:
                errors += [msg_template.format(inp["id"])]
            else:
                errors += []

            msg_template = (
                ' InputError: "{0}" cannot have default' " value outside its choices"
            )
            if "default-value" in inp.keys():
                if not isinstance(inp["default-value"], list):
                    if inp["default-value"] not in inp["value-choices"]:
                        errors += [msg_template.format(inp["id"])]
                else:
                    for dv in inp["default-value"]:
                        if dv not in inp["value-choices"]:
                            errors += [msg_template.format(inp["id"])]
            else:
                errors += []

        # Verify list-type inputs (min entries less than max,
        # no negative entries (both on min and max)
        if "list" in inp.keys():
            msg_template = (
                ' InputError: "{0}" cannot have greater min'
                " entries ({1}) than max entries ({2})"
            )
            minn = inp.get("min-list-entries") or 0
            maxx = inp.get("max-list-entries") or float("Inf")
            if minn > maxx:
                errors += [msg_template.format(inp["id"], minn, maxx)]
            else:
                errors += []

            msg_template = (
                ' InputError: "{0}" cannot have negative min' " entries ({1})"
            )
            errors += [msg_template.format(inp["id"], minn)] if minn < 0 else []

            msg_template = (
                ' InputError: "{0}" cannot have non-positive' " max entries ({1})"
            )
            if maxx <= 0:
                errors += [msg_template.format(inp["id"], maxx)]
            else:
                errors += []

        # Verify requires-inputs (present ids, non-overlapping)
        msg_template = ' InputError: "{0}" required id "{1}" not found'
        if "requires-inputs" in inp.keys():
            errors += [
                msg_template.format(inp["id"], ids)
                for ids in inp["requires-inputs"]
                if ids not in inIds + grpIds
            ]

        # Verify disables-inputs (present ids, non-overlapping)
        msg_template = ' InputError: "{0}" disables id "{1}" not found'
        if "disables-inputs" in inp.keys():
            errors += [
                msg_template.format(inp["id"], ids)
                for ids in inp["disables-inputs"]
                if ids not in inIds
            ]

        if "requires-inputs" in inp.keys() and "disables-inputs" in inp.keys():
            msg_template = ' InputError: "{0}" requires and disables "{1}"'
            errors += [
                msg_template.format(inp["id"], ids1)
                for ids1 in inp["requires-inputs"]
                for ids2 in inp["disables-inputs"]
                if ids1 == ids2
            ]

        # Verify that disableed inputs cannot be required
        msg_template = ' InputError: "{0}" disables required id "{1}"'
        if "disables-inputs" in inp.keys():
            errors += [
                msg_template.format(inp["id"], ids)
                for ids in inp["disables-inputs"]
                if ids in inIds and not inById(ids).get("optional")
            ]

        # Verify required inputs cannot require or disable other parameters
        if "requires-inputs" in inp.keys() or "disables-inputs" in inp.keys():
            msg_template = (
                ' InputError: "{0}" cannot require or' " disable other inputs"
            )
            if not inp["optional"]:
                errors += [msg_template.format(inp["id"])]

        # Verify value-disables/requires fields accompany value-choices
        if (
            "value-disables" in inp.keys() or "value-requires" in inp.keys()
        ) and "value-choices" not in inp.keys():
            msg_template = (
                ' InputError: "{0}" cannot have have value-opts'
                " without value-choices defined."
            )
            errors += [msg_template.format(inp["id"])]

        if "value-choices" in inp.keys():
            # Verify not value not requiring and disabling input
            if "value-requires" in inp.keys() and "value-disables" in inp.keys():
                msg_template = (
                    ' InputError: "{0}" choice "{1}" requires' ' and disables "{2}"'
                )
                errors += [
                    msg_template.format(inp["id"], choice, ids1)
                    for choice in inp["value-choices"]
                    for ids1 in inp["value-disables"][choice]
                    if ids1 in inp["value-requires"][choice]
                ]

            for param in ["value-requires", "value-disables"]:
                if param in inp.keys():
                    # Verify disables/requires keys are the same as choices
                    msg_template = (
                        ' InputError: "{0}" {1} list is not the'
                        " same as the value-choices"
                    )
                    if set(inp[param].keys()) != set(inp["value-choices"]):
                        errors += [msg_template.format(inp["id"], param)]

                    # Verify all required or disabled IDs are valid
                    msg_template = ' InputError: "{0}" {1} id "{2}" not' " found"
                    errors += [
                        msg_template.format(inp["id"], param, ids)
                        for ids in inp[param].values()
                        for item in ids
                        if item not in inIds
                    ]

                    # Verify not requiring or disabling required inputs
                    msg_template = (
                        ' InputError: "{0}" {1} cannot be used '
                        'with required input "{2}"'
                    )
                    errors += [
                        msg_template.format(inp["id"], param, member)
                        for ids in inp[param].keys()
                        for member in inp[param][ids]
                        if not inById(member).get("optional")
                    ]

    # Verify groups
    for idx, grpid in enumerate(grpIds):
        grp = descriptor["groups"][idx]
        # Verify group members must (exist in inputs, show up
        # once, only belong to single group)
        msg_template = ' GroupError: "{0}" member "{1}" does not exist'
        errors += [
            msg_template.format(grp["id"], member)
            for member in grp["members"]
            if member not in inIds
        ]

        msg_template = ' GroupError: "{0}" member "{1}" appears twice'
        errors += [
            msg_template.format(grp["id"], member)
            for member in set(grp["members"])
            if grp["members"].count(member) > 1
        ]

        # Verify mutually exclusive groups cannot have required members
        # nor requiring members, and that pairs of inputs cannot both be
        # in an all-or-none group
        if grp.get("mutually-exclusive"):
            msg_template = (
                ' GroupError: "{0}" is mutually-exclusive'
                " and cannot have required members, "
                'such as "{1}"'
            )
            errors += [
                msg_template.format(grp["id"], member)
                for member in set(grp["members"])
                if not inById(member)["optional"]
            ]

            msg_template = (
                ' GroupError: "{0}" is mutually-exclusive'
                " and cannot have members require one another,"
                ' such as "{1}" and "{2}"'
            )
            for member in set(grp["members"]):
                if "requires-inputs" in inById(member).keys():
                    errors += [
                        msg_template.format(grp["id"], member, req)
                        for req in inById(member)["requires-inputs"]
                        if req in set(grp["members"])
                    ]

            for jdx, grp2 in enumerate(descriptor["groups"]):
                if grp2.get("all-or-none"):
                    msg_template = (
                        " GroupError: mutually-exclusive group"
                        ' "{0}" and all-or-none group "{1}"'
                        ' cannot both contain input pairs "{2}"'
                        ' and "{3}"'
                    )
                    errors += [
                        msg_template.format(grp["id"], grp2["id"], m1, m2)
                        for m1 in grp["members"]
                        for m2 in grp["members"]
                        if m1 != m2
                        and m1 in grp2["members"]
                        and m2 in grp2["members"]
                        and idx != jdx
                    ]

        # Verify one-is-required groups should never have required members
        # and that the group is not a subset of an all-or-none group
        if grp.get("one-is-required"):
            msg_template = (
                ' GroupError: "{0}" is a one-is-required'
                ' group and contains a required member, "{1}"'
            )
            errors += [
                msg_template.format(grp["id"], member)
                for member in set(grp["members"])
                if member in inIds and not inById(member)["optional"]
            ]

            for jdx, grp2 in enumerate(descriptor["groups"]):
                if grp2.get("all-or-none"):
                    msg_template = (
                        ' GroupError: "{0}" is one-is-required'
                        " and cannot be a subset of the all-or-none"
                        ' group "{1}"'
                    )
                    if (
                        set(grp["members"]).issubset(set(grp2["members"]))
                        and idx != jdx
                    ):
                        errors += [msg_template.format(grp["id"], grp2["id"])]

        # Verify all-or-none groups should never have required members
        if grp.get("all-or-none"):
            msg_template = (
                ' GroupError: "{0}" is an all-or-none group'
                " and cannot be paired with one-is-required"
                " or mutually-exclusive groups"
            )
            if grp.get("one-is-required") or grp.get("mutually-exclusive"):
                errors += [msg_template.format(grp["id"])]

            msg_template = (
                ' GroupError: "{0}" is an all-or-none'
                ' group and contains a required member, "{1}"'
            )
            errors += [
                msg_template.format(grp["id"], member)
                for member in set(grp["members"])
                if member in inIds and not inById(member)["optional"]
            ]

    # Verify tests
    if "tests" in descriptor.keys():
        tests_names = []
        for test in descriptor["tests"]:

            tests_names.append(test["name"])
            if "output-files" in test["assertions"].keys():
                test_output_ids = safeGet(test["assertions"], "output-files", "id")

                # Verify if output reference ids are valid
                msg_template = 'TestError: "{0}" output id' ' not found, in test "{1}"'
                errors += [
                    msg_template.format(output_id, test["name"])
                    for output_id in test_output_ids
                    if (output_id not in outIds)
                ]

                # Verify that we do not have multiple output
                # references referring to the same id
                msg_template = (
                    'TestError: "{0}" output id'
                    " cannot appear more than once within"
                    ' same test, in test "{1}"'
                )
                errors += [
                    msg_template.format(output_id, test["name"])
                    for output_id in set(test_output_ids)
                    if (test_output_ids.count(output_id) > 1)
                ]

        # Verify that all the defined tests have unique names
        msg_template = 'TestError: "{0}" test name is non-unique'
        errors += [
            msg_template.format(test_name)
            for test_name in set(tests_names)
            if (tests_names.count(test_name) > 1)
        ]

    # Verify container image
    if "container-image" in descriptor.keys():
        conImage = descriptor["container-image"]["image"]
        conIndex = (
            descriptor["container-image"]["index"]
            if "index" in descriptor["container-image"]
            else None
        )
        conImageIndex = re.match(r"^[a-zA-Z0-9]+://", conImage)

        # Verify index in container image is equal to container index value
        if (
            conIndex is not None
            and conImageIndex is not None
            and conIndex != conImageIndex.group()
        ):
            msg_template = (
                'ContainerError: container image "{0}"'
                " is prepended by index that doesn't match"
                ' container index value "{1}"'
            )
            errors += [msg_template.format(conImage, conIndex)]

    errors = None if errors == [] else errors
    if errors is None:
        if kwargs.get("format_output") and kwargs.get("descriptor_path"):
            with open(kwargs.get("descriptor_path"), "w") as fhandle:
                fhandle.write(
                    json.dumps(customSortDescriptorByKey(descriptor), indent=4)
                )
        return descriptor
    else:
        raise DescriptorValidationError("\n".join(errors))



#!/usr/bin/env python

import os
import subprocess

import pytest

from boutiques import __file__ as bfile
from boutiques.bosh import bosh
from boutiques.tests.BaseTest import BaseTest


class TestInvocation(BaseTest):
    @pytest.fixture(autouse=True)
    def set_test_dir(self):
        self.setup("invocation")

    def test_invocation(self):
        descriptor = self.get_file_path("good.json")
        invocation = self.get_file_path("good_invocation.json")
        self.assertFalse(bosh(["invocation", descriptor, "-i", invocation, "-w"]))
        self.assertFalse(bosh(["invocation", descriptor, "-i", invocation, "-w"]))

    def test_invocation_json_obj(self):
        descriptor = open(self.get_file_path("good.json")).read()
        invocation = open(self.get_file_path("good_invocation.json")).read()
        self.assertFalse(bosh(["invocation", descriptor, "-i", invocation, "-w"]))
        self.assertFalse(bosh(["invocation", descriptor, "-i", invocation, "-w"]))

    def test_invocation_invalid_cli(self):
        descriptor = self.get_file_path("good.json")
        invocation = self.get_file_path("wrong_invocation.json")
        command = "bosh invocation " + descriptor + "-i " + invocation
        process = subprocess.Popen(
            command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        process.communicate()
        self.assertTrue(process.returncode)



import os.path as op
import subprocess
import tempfile

import pytest
import simplejson as json

from boutiques.util.utils import loadJson


def pytest_addoption(parser):
    parser.addoption("--descriptor", action="append", default=[])

    # Additional options during bosh exec launch:
    parser.addoption("--imagepath", action="store")


def fetch_tests(descriptor_input, paramsDict):
    descriptor = loadJson(descriptor_input)

    tests = []

    # For each test present in the descriptor:
    for test in descriptor["tests"]:

        # We first extract the invocation and put it inside a temporary file.
        invocation_JSON = json.dumps(test["invocation"])
        temp_invocation_JSON = tempfile.NamedTemporaryFile(suffix=".json", delete=False)
        temp_invocation_JSON.write(invocation_JSON.encode())
        temp_invocation_JSON.seek(0)

        # Now we setup the necessary elements for the testing function.
        tests.append([descriptor_input, test, temp_invocation_JSON, paramsDict])

    return (descriptor["name"], tests)


# This function will be executed by pytest before the actual testing
def pytest_generate_tests(metafunc):
    descriptor_filename = metafunc.config.getoption("descriptor")[0]
    additional_params = {"--skip-data-collection": None}

    if metafunc.config.option.imagepath is not None:
        additional_params["--imagepath"] = metafunc.config.option.imagepath
    # Each element in 'tests' will hold the necessary information
    # for a single test
    # Those information are:
    #     . The descriptor (common to all)
    #     . The related JSON data, describing the test
    #       (more convenient, no need to extract
    #        again from descriptor)
    #     . The invocation file needed for the test
    #     . Any additional options listed in conftest.pytest_addoption()
    descriptor_name, tests = fetch_tests(descriptor_filename, additional_params)

    # Generate the test ids for each of the test cases.
    # An id is created by concatenaning the name of the descriptor
    # with the name of the test case.
    names = [
        f"{op.basename(descriptor_filename)}_{params[1]['name'].replace(' ', '-')}"
        for params in tests
    ]

    metafunc.parametrize("descriptor, test, invocation, paramsDict", tests, ids=names)


############
# Fixtures #
############
@pytest.fixture
def skip_if_no_apptainer():
    if subprocess.Popen("type apptainer", shell=True).wait():
        pytest.skip("Apptainer not installed")


@pytest.fixture
def skip_if_no_docker():
    if subprocess.Popen("type docker", shell=True).wait():
        pytest.skip("Docker not installed")



"""
BOUTIQUES COMMANDS

TOOL CREATION
* create: create a Boutiques descriptor from scratch.
* export: export a descriptor to other formats.
* import: create a descriptor for a BIDS app or update a descriptor from \
an older version of the schema. Options: "bids", "0.4", "cwl", "docopt"
* validate: validate an existing boutiques descriptor.

TOOL USAGE & EXECUTION
* example: generate example command-line for descriptor.
* pprint: generate pretty help text from a descriptor.
* exec: launch or simulate an execution given a descriptor and a set of inputs.
* test: run pytest on a descriptor detailing tests.

TOOL SEARCH & PUBLICATION
* deprecate: deprecate a published tool. The tool will still be published and
usable, but it won't show in search results.
* publish: create an entry in Zenodo for the descriptor and adds the DOI \
created by Zenodo to the descriptor.
* pull: download a descriptor from Zenodo.
* search: search Zenodo for descriptors.

DATA COLLECTION
* data: manage execution data collection.

OTHER
* evaluate: given an invocation and a descriptor,queries execution properties.
* invocation: generate or validate inputs against the invocation schema
* for a given descriptor.
* version: print the Boutiques version.
"""

import os
from argparse import ArgumentParser, RawTextHelpFormatter

import simplejson as json


def add_subparser_create(subparsers):
    parser_create = subparsers.add_parser(
        "create", description="Boutiques descriptor creator"
    )
    parser_create.set_defaults(function="create")
    parser_create.add_argument(
        "descriptor",
        action="store",
        help="Output file to store descriptor in.",
    )
    parser_create.add_argument(
        "--docker-image",
        "-d",
        action="store",
        help="Name of Docker image on DockerHub.",
    )
    parser_create.add_argument(
        "--use-singularity",
        "-u",
        action="store_true",
        help="When --docker-image is used. Specify to " "use singularity to run it.",
    )
    parser_create.add_argument(
        "--cl-template",
        action="store",
        help="Command-line template used to generate"
        " descriptor inputs. (Path to descriptor or String)",
    )
    parser_create.add_argument(
        "--camel-case",
        action="store_true",
        help="All input IDs will be written in camelCase.",
    )


def add_subparser_data(subparsers):
    parser_data = subparsers.add_parser(
        "data", description="Manage execution data collection."
    )
    parser_data.set_defaults(function="data")
    data_subparsers = parser_data.add_subparsers(
        help="Manage execution data records. Inspect: displays "
        "the unpublished records currently in the cache. "
        "Publish: publishes contents of cache to Zenodo as "
        "a public data set. Requires a Zenodo access token, "
        "see http://developers.zenodo.org/#authentication. "
        "Delete: remove one or more records from the cache."
    )

    parser_data_delete = data_subparsers.add_parser(
        "delete", description="Delete data record(s) in cache."
    )
    parser_data_delete.set_defaults(mode="delete")
    parser_data_delete.add_argument(
        "-f", "--file", action="store", help="Filename of record to delete."
    )
    parser_data_delete.add_argument(
        "--no-int",
        "-y",
        action="store_true",
        help="disable interactive input.",
    )

    parser_data_inspect = data_subparsers.add_parser(
        "inspect", description="Displays contents of cache"
    )
    parser_data_inspect.set_defaults(mode="inspect")
    parser_data_inspect.add_argument(
        "-e",
        "--example",
        action="store_true",
        help="Display example data file contents.",
    )
    parser_data_inspect.add_argument(
        "-l",
        "--latest",
        action="store_true",
        help="Display latest data file contents.",
    )

    parser_data_publish = data_subparsers.add_parser(
        "publish", description="Publishes record(s) to a data set."
    )
    parser_data_publish.set_defaults(mode="publish")
    parser_data_publish.add_argument(
        "-a",
        "--author",
        action="store",
        help="Set the author name for the data set "
        "publication. Defaults to anonymous.",
    )
    parser_data_publish.add_argument(
        "-f",
        "--file",
        action="store",
        help="Filename of record to publish alone as a " "data set.",
    )
    parser_data_publish.add_argument(
        "-i",
        "--individually",
        action="store_true",
        help="Publishes all data files in cache as "
        "independent data sets, By Default will publish "
        "files in bulk data sets.",
    )
    parser_data_publish.add_argument(
        "--no-int",
        "-y",
        action="store_true",
        help="disable interactive input.",
    )
    parser_data_publish.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="print information messages.",
    )
    parser_data_publish.add_argument(
        "--sandbox",
        action="store_true",
        help="publish to Zenodo's sandbox instead of "
        "production server. Recommended for tests.",
    )
    parser_data_publish.add_argument(
        "--zenodo-token",
        action="store",
        help="Zenodo API token to use for authentication. "
        "If not used, token will be read from "
        "configuration file or requested interactively.",
    )
    parser_data_publish.add_argument(
        "--nexus",
        action="store_true",
        help="Publish to Nexus instead of Zenodo. "
        "Sandbox URL is "
        "https://sandbox.bluebrainnexus.io",
    )
    parser_data_publish.add_argument(
        "--nexus-token",
        action="store",
        help="Nexus API token to use for authentication. ",
    )
    parser_data_publish.add_argument(
        "--nexus-org",
        action="store",
        help="Nexus organization to publish to. ",
    )
    parser_data_publish.add_argument(
        "--nexus-project", action="store", help="Nexus project to publish to. "
    )


def add_subparser_deprecate(subparsers):
    parser_deprecate = subparsers.add_parser(
        "deprecate",
        description="Deprecates a published descriptor by"
        " creating a new version with the 'deprecated' tag"
        " on Zenodo. The descriptor remains available from"
        " its Zenodo id, but it won't show in search"
        " results. This works by creating a new version of"
        " the tool in Zenodo, marked with keyword"
        " 'deprecated'.",
    )
    parser_deprecate.set_defaults(function="deprecate")
    parser_deprecate.add_argument(
        "zid",
        action="store",
        help="Zenodo id "
        "of the descriptor to deprecate, "
        "prefixed by 'zenodo.', e.g. zenodo.123456",
    )
    parser_deprecate.add_argument(
        "--by",
        action="store",
        help="Zenodo id (e.g., "
        "zenodo-1234) of a  descriptor that will supersede "
        "the deprecated one.",
    )
    parser_deprecate.add_argument(
        "--zenodo-token",
        action="store",
        help="Zenodo API token to use for authentication. "
        "If not used, token will be read from configuration "
        "file or requested interactively.",
    )
    parser_deprecate.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Print information messages",
    )
    parser_deprecate.add_argument(
        "--sandbox",
        action="store_true",
        help="use Zenodo's sandbox instead of "
        "production server. Recommended for tests.",
    )


def add_subparser_evaluate(subparsers):
    parser_evaluate = subparsers.add_parser(
        "evaluate",
        description="Evaluates parameter values for a" " descriptor and invocation",
    )
    parser_evaluate.set_defaults(function="evaluate")
    parser_evaluate.add_argument(
        "descriptor",
        action="store",
        help="The Boutiques descriptor as a JSON file, JSON "
        "string or Zenodo ID (prefixed by 'zenodo.').",
    )
    parser_evaluate.add_argument(
        "invocation",
        action="store",
        help="Input JSON complying to invocation.",
    )
    parser_evaluate.add_argument(
        "query",
        action="store",
        nargs="*",
        help="The query to be performed. Simply request keys "
        "from the descriptor (i.e. output-files), and chain "
        "together queries (i.e. id=myfile or optional=false) "
        "slashes between them and commas connecting them. "
        "(i.e. output-files/optional=false,id=myfile). "
        "Perform multiple queries by separating them with a "
        "space.",
    )
    parser_evaluate.add_argument(
        "--sandbox",
        action="store_true",
        help="Get descriptor from Zenodo's sandbox instead of " "production server.",
    )


def add_subparser_example(subparsers):
    parser_example = subparsers.add_parser(
        "example",
        description="Generates example invocation from a" " valid descriptor",
    )
    parser_example.set_defaults(function="example")
    parser_example.add_argument(
        "descriptor",
        action="store",
        help="The Boutiques descriptor as a JSON file, "
        "JSON string or Zenodo ID (prefixed by 'zenodo.').",
    )
    parser_example.add_argument(
        "-c",
        "--complete",
        action="store_true",
        help="Include optional parameters.",
    )
    parser_example.add_argument(
        "--sandbox",
        action="store_true",
        help="Get descriptor from Zenodo's sandbox instead of " "production server.",
    )


def add_subparser_execute(subparsers):
    parser_exec = subparsers.add_parser(
        "exec", description="Boutiques local executor", aliases=["execute"]
    )
    parser_exec.set_defaults(function="exec")
    exec_subparsers = parser_exec.add_subparsers(
        help="Mode of operation to use. Launch: takes a "
        "set of inputs compliant with invocation schema "
        "and launches the tool. Simulate: shows sample "
        "command-lines based on the provided descriptor"
        " based on provided or randomly generated inputs. "
        "Prepare: pulls the Docker or Singularity container "
        "image for a given descriptor. "
    )

    parser_exec_launch = exec_subparsers.add_parser(
        "launch", description="Launches an invocation."
    )
    parser_exec_launch.set_defaults(mode="launch")
    parser_exec_launch.add_argument(
        "descriptor",
        action="store",
        help="The Boutiques descriptor as a JSON file, "
        "JSON string or Zenodo ID (prefixed by 'zenodo.').",
    )
    parser_exec_launch.add_argument(
        "invocation",
        action="store",
        help="Input JSON complying to invocation.",
    )
    parser_exec_launch.add_argument(
        "-v",
        "--volumes",
        action="append",
        type=str,
        help="Volumes to mount when launching the "
        "container. Format consistently the following:"
        " /a:/b will mount local directory /a to "
        "container directory /b.",
    )
    parser_exec_launch.add_argument(
        "--container-opts",
        action="append",
        type=str,
        help="Options to pass to the container runtime. Overrides the descriptor's container options.",
    )
    parser_exec_launch.add_argument(
        "-x",
        "--debug",
        action="store_true",
        help="Keeps temporary scripts used during "
        "execution, and prints additional debug "
        "messages.",
    )
    parser_exec_launch.add_argument(
        "-u",
        "--user",
        action="store_true",
        help=f"Runs the container as local user ({os.getenv('USER')}) instead of root.",
    )
    parser_exec_launch.add_argument(
        "-s",
        "--stream",
        action="store_true",
        help="Streams stdout and stderr in real time " "during execution.",
    )
    parser_exec_launch.add_argument(
        "--imagepath",
        action="store",
        help="Path to Singularity image. "
        "If not specified, will use current directory.",
    )
    parser_exec_launch.add_argument(
        "--skip-data-collection",
        action="store_true",
        help="Skips execution data collection and saving" "to cache.",
    )
    parser_exec_launch.add_argument(
        "--provenance",
        action="store",
        type=json.loads,
        help="Append JSON object to the generated record.",
    )
    parser_exec_launch.add_argument(
        "--no-container",
        action="store_true",
        help="Launch invocation on the host computer, with "
        "no container. If 'container-image' appears in the "
        "descriptor, it is ignored.",
    )
    parser_exec_launch.add_argument(
        "--sandbox",
        action="store_true",
        help="Get descriptor from Zenodo's sandbox instead of " "production server.",
    )
    parser_exec_launch.add_argument(
        "--no-automounts",
        action="store_true",
        help="Disable automatic mount of all input files " "present in the invocation",
    )
    force_group = parser_exec_launch.add_mutually_exclusive_group()
    force_group.add_argument(
        "--force-docker",
        action="store_true",
        help="Tries to run Singularity images with "
        "Docker. This only works if the image is on"
        "Docker Hub, i.e. has index docker://",
    )
    force_group.add_argument(
        "--force-singularity",
        action="store_true",
        help="Tries to run Docker images with " "Singularity.",
    )

    parser_exec_prepare = exec_subparsers.add_parser(
        "prepare",
        description="Pulls the container image for a given descriptor",
    )
    parser_exec_prepare.set_defaults(mode="prepare")
    parser_exec_prepare.add_argument(
        "descriptor",
        action="store",
        help="The Boutiques descriptor as a JSON file, "
        "JSON string or Zenodo ID (prefixed by 'zenodo.').",
    )
    parser_exec_prepare.add_argument(
        "-x",
        "--debug",
        action="store_true",
        help="Keeps temporary scripts used during "
        "execution, and prints additional debug "
        "messages.",
    )
    parser_exec_prepare.add_argument(
        "-s",
        "--stream",
        action="store_true",
        help="Streams stdout and stderr in real time " "during execution.",
    )
    parser_exec_prepare.add_argument(
        "--imagepath",
        action="store",
        help="Path to Singularity image. "
        "If not specified, will use current directory.",
    )
    parser_exec_prepare.add_argument(
        "--sandbox",
        action="store_true",
        help="Get descriptor from Zenodo's sandbox instead of " "production server.",
    )

    parser_exec_simulate = exec_subparsers.add_parser(
        "simulate", description="Simulates an invocation."
    )
    parser_exec_simulate.set_defaults(mode="simulate")
    parser_exec_simulate.add_argument(
        "descriptor",
        action="store",
        help="The Boutiques descriptor as a JSON file, "
        "JSON string or Zenodo ID (prefixed by 'zenodo.').",
    )
    parser_exec_simulate.add_argument(
        "-i",
        "--input",
        action="store",
        help="Input JSON complying to invocation.",
    )
    parser_exec_simulate.add_argument(
        "-j",
        "--json",
        action="store_true",
        help="Flag to generate invocation in JSON format.",
    )
    parser_exec_simulate.add_argument(
        "-c",
        "--complete",
        action="store_true",
        help="Include optional parameters.",
    )
    parser_exec_simulate.add_argument(
        "--sandbox",
        action="store_true",
        help="Get descriptor from Zenodo's sandbox instead of " "production server.",
    )


def add_subparser_export(subparsers):
    parser_export = subparsers.add_parser(
        "export",
        description="Export Boutiques descriptor to" " other formats.",
    )
    parser_export.set_defaults(function="export")
    parser_export.add_argument(
        "type", help="Type of export we are performing.", choices=["carmin"]
    )
    parser_export.add_argument("descriptor", help="Boutiques descriptor to export.")
    parser_export.add_argument(
        "--identifier", help="Identifier to use in" "CARMIN export."
    )
    parser_export.add_argument(
        "--sandbox",
        action="store_true",
        help="Get descriptor from Zenodo's sandbox instead of " "production server.",
    )
    parser_export.add_argument(
        "output",
        help="Output file where to write the" " converted descriptor.",
    )


def add_subparser_import(subparsers):
    parser_import = subparsers.add_parser(
        "import",
        description="Imports old descriptor or BIDS app or"
        " CWL descriptor or docopt script to spec.",
    )
    parser_import.set_defaults(function="import")
    parser_import.add_argument(
        "type",
        help="Type of import we are performing. Allowed values: {}".format(
            ", ".join(["bids", "0.4", "cwl", "docopt", "config"])
        ),
        choices=["bids", "0.4", "cwl", "docopt", "config"],
        metavar="type",
    )
    parser_import.add_argument(
        "output_descriptor",
        help="Where the Boutiques" " descriptor will be written.",
    )
    parser_import.add_argument(
        "input_descriptor",
        help="Input descriptor to be"
        " converted. For '0.4', is JSON descriptor,"
        " for 'docopt' is Docopt script,"
        " for 'bids' is base directory of BIDS app,"
        " for 'cwl' is YAML descriptor."
        " for 'config' is configuration file"
        " {.json, .toml, .yml}.)",
    )
    parser_import.add_argument(
        "-o",
        "--output-invocation",
        help="Where to write " "the invocation if any.",
    )
    parser_import.add_argument(
        "-i", "--input-invocation", help="Input invocation " " for CWL if any."
    )


def add_subparser_invocation(subparsers):
    parser_invocation = subparsers.add_parser(
        "invocation",
        description="Creates invocation schema and"
        " validates invocations. Uses descriptor's"
        " invocation schema if it exists, otherwise"
        " creates one.",
    )
    parser_invocation.set_defaults(function="invocation")
    parser_invocation.add_argument(
        "descriptor",
        action="store",
        help="The Boutiques descriptor as a JSON file, JSON "
        "string or Zenodo ID (prefixed by 'zenodo.').",
    )
    parser_invocation.add_argument(
        "-i",
        "--invocation",
        action="store",
        help="Input values in a JSON file or as a JSON "
        "object to be validated against "
        "the invocation schema.",
    )
    parser_invocation.add_argument(
        "-w",
        "--write-schema",
        action="store_true",
        help="If descriptor doesn't have an invocation "
        "schema, creates one and writes it to the descriptor"
        " file ",
    )
    parser_invocation.add_argument(
        "--sandbox",
        action="store_true",
        help="Get descriptor from Zenodo's sandbox instead of " "production server.",
    )


def add_subparser_pprint(subparsers):
    parser_pprint = subparsers.add_parser(
        "pprint",
        description="Boutiques pretty-print for" "generating help text",
    )
    parser_pprint.set_defaults(function="pprint")
    parser_pprint.add_argument(
        "descriptor", action="store", help="The Boutiques descriptor."
    )
    parser_pprint.add_argument(
        "--sandbox",
        action="store_true",
        help="Get descriptor from Zenodo's sandbox instead of " "production server.",
    )


def add_subparser_publish(subparsers):
    parser_publish = subparsers.add_parser(
        "publish",
        description="A publisher of Boutiques tools"
        " in Zenodo (http://zenodo.org). Requires "
        "a Zenodo access token, see "
        "http://developers.zenodo.org/#authentication.",
    )
    parser_publish.set_defaults(function="publish")
    parser_publish.add_argument(
        "boutiques_descriptor",
        action="store",
        help="local path of the " " Boutiques descriptor to publish.",
    )
    parser_publish.add_argument(
        "--sandbox",
        action="store_true",
        help="publish to Zenodo's sandbox instead of "
        "production server. Recommended for tests.",
    )
    parser_publish.add_argument(
        "--zenodo-token",
        action="store",
        help="Zenodo API token to use for authentication. "
        "If not used, token will be read from configuration "
        "file or requested interactively.",
    )
    parser_publish.add_argument(
        "--no-int",
        "-y",
        action="store_true",
        help="disable interactive input.",
    )
    parser_publish.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="print information messages.",
    )
    publish_mutex_group = parser_publish.add_mutually_exclusive_group()
    publish_mutex_group.add_argument(
        "-r",
        "--replace",
        action="store_true",
        help="Publish an updated version of an existing "
        "record. The descriptor must contain a DOI, which "
        "will be replaced with a new one.",
    )
    publish_mutex_group.add_argument(
        "--id",
        action="store",
        help="Zenodo ID of an existing record you wish to "
        "update with a new version, prefixed by "
        "'zenodo.' (e.g. zenodo.123456).",
    )


def add_subparser_pull(subparsers):
    parser_pull = subparsers.add_parser(
        "pull",
        description="Ensures that Zenodo descriptors are"
        " locally cached, downloading them if needed.",
    )
    parser_pull.set_defaults(function="pull")
    parser_pull.add_argument(
        "zids",
        nargs="+",
        action="store",
        help="One or "
        "more Zenodo IDs for the descriptor(s) to pull, "
        "prefixed by 'zenodo.', e.g. zenodo.123456 "
        "zenodo.123457",
    )
    parser_pull.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Print information messages",
    )
    parser_pull.add_argument(
        "--sandbox",
        action="store_true",
        help="pull from Zenodo's sandbox instead of "
        "production server. Recommended for tests.",
    )


def add_subparser_search(subparsers):
    parser_search = subparsers.add_parser(
        "search",
        description="Search Zenodo for Boutiques"
        " descriptors. When no term is supplied, will"
        " search for all descriptors.",
    )
    parser_search.set_defaults(function="search")
    parser_search.add_argument(
        "query",
        nargs="?",
        default="boutiques",
        action="store",
        help="Search query",
    )
    parser_search.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Print information messages",
    )
    parser_search.add_argument(
        "--sandbox",
        action="store_true",
        help="search Zenodo's sandbox instead of "
        "production server. Recommended for tests.",
    )
    parser_search.add_argument(
        "-m",
        "--max",
        action="store",
        type=int,
        help="Specify the maximum number of results " "to be returned. Default is 10.",
    )
    parser_search.add_argument(
        "-nt",
        "--no-trunc",
        action="store_true",
        help="Do not truncate long tool descriptions.",
    )
    parser_search.add_argument(
        "-e",
        "--exact",
        action="store_true",
        help="Only return results containing the exact query.",
    )


def add_subparser_test(subparsers):
    parser_test = subparsers.add_parser(
        "test",
        description="Perform all the tests defined within" " the given descriptor",
    )
    parser_test.set_defaults(function="test")
    parser_test.add_argument(
        "descriptor",
        action="store",
        help="The Boutiques descriptor as a JSON file, JSON "
        "string or Zenodo ID (prefixed by 'zenodo.').",
    )
    parser_test.add_argument(
        "--sandbox",
        action="store_true",
        help="Get descriptor from Zenodo's sandbox instead of " "production server.",
    )
    parser_test.add_argument(
        "--imagepath",
        action="store",
        help="Path to Singularity image. "
        "If not specified, will use current directory.",
    )


def add_subparser_validate(subparsers):
    parser_validate = subparsers.add_parser(
        "validate", description="Boutiques descriptor validator"
    )
    parser_validate.set_defaults(function="validate")
    parser_validate.add_argument(
        "descriptor",
        action="store",
        help="The Boutiques descriptor as a JSON file, JSON "
        "string or Zenodo ID (prefixed by 'zenodo.').",
    )
    parser_validate.add_argument(
        "--bids",
        "-b",
        action="store_true",
        help="Flag indicating if descriptor is a BIDS app",
    )
    parser_validate.add_argument(
        "--format",
        "-f",
        action="store_true",
        help="If descriptor is valid, rewrite it with sorted" " keys.",
    )
    parser_validate.add_argument(
        "--sandbox",
        action="store_true",
        help="Get descriptor from Zenodo's sandbox instead of " "production server.",
    )


def add_subparser_version(subparsers):
    parser_version = subparsers.add_parser(
        "version", description="Print the Boutiques version."
    )
    parser_version.set_defaults(function="version")


def parser_bosh():
    parser = ArgumentParser(add_help=False, formatter_class=RawTextHelpFormatter)
    parser.add_argument(
        "--help",
        "-h",
        action="store_true",
        help="show this help message and exit",
    )
    subparsers = parser.add_subparsers(help=__doc__)
    add_subparser_create(subparsers)
    add_subparser_data(subparsers)
    add_subparser_deprecate(subparsers)
    add_subparser_evaluate(subparsers)
    add_subparser_example(subparsers)
    add_subparser_execute(subparsers)
    add_subparser_export(subparsers)
    add_subparser_import(subparsers)
    add_subparser_invocation(subparsers)
    add_subparser_pprint(subparsers)
    add_subparser_publish(subparsers)
    add_subparser_pull(subparsers)
    add_subparser_search(subparsers)
    add_subparser_test(subparsers)
    add_subparser_validate(subparsers)
    add_subparser_version(subparsers)
    return parser



#!/usr/bin/env python

import subprocess

import pytest

from boutiques.tests.BaseTest import BaseTest


class TestExample(BaseTest):
    @pytest.fixture(autouse=True)
    def set_test_dir(self):
        self.setup("import/docopt/")

    def test_reqInpGroup_nf_wGroup_OK(self):
        nf_desc = self.get_file_path("naval_fate.json")
        invocation = self.get_file_path("nf_invoc_shoot.json")
        command = "bosh invocation " + nf_desc + " -i " + invocation
        process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE)
        self.assertEqual(process.stdout.read().strip(), b"OK")

    def test_reqInpGroup_nf_wOptionalFlag_OK(self):
        nf_desc = self.get_file_path("naval_fate.json")
        invocation = self.get_file_path("nf_invoc_move.json")
        command = "bosh invocation " + nf_desc + " -i " + invocation
        process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE)
        self.assertEqual(process.stdout.read().strip(), b"OK")

    def test_reqInpGroup_nf_wChildren_woGroup_FAIL(self):
        nf_desc = self.get_file_path("naval_fate.json")
        invocation = self.get_file_path("nf_invoc_missing_group.json")
        command = "bosh invocation " + nf_desc + " -i " + invocation
        process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE)
        self.assertIn(b"Failed validating", process.stdout.read().strip())

    def test_reqInpGroup_nf_woChildren_woGroup_FAIL(self):
        nf_desc = self.get_file_path("naval_fate.json")
        invocation = self.get_file_path("nf_invoc_missing_all.json")
        command = "bosh invocation " + nf_desc + " -i " + invocation
        process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE)
        self.assertIn(b"Failed validating", process.stdout.read().strip())

    def test_reqInpGroup_valid_Complex_OK(self):
        valid_desc = self.get_file_path("test_valid.json")
        invocation = self.get_file_path("valid_invoc_mutex.json")
        command = "bosh invocation " + valid_desc + " -i " + invocation
        process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE)
        self.assertEqual(process.stdout.read().strip(), b"OK")



#!/usr/bin/env python

import argparse
import os
import sys


def file_exists(parser, file_name):
    if not os.path.exists(file_name):
        parser.error(f"File not found: {file_name}")
    return file_name


def is_valid_enum_value(parser, value):
    if value not in ["val1", "val2", "val3"]:
        parser.error(f"Invalid enum value: {value}")
    return value


def test_mutex_group(results):
    if results.number and results.enum_input:
        sys.stderr.write("error: number and enum_input are mutually exclusive.\n")
        sys.exit(1)


def test_one_is_required_group(results):
    if (not results.number) and (not results.enum_input):
        sys.stderr.write("error: one of number or enum_input is required.\n")
        sys.exit(1)


def test_envVar(results):
    var = os.getenv("ENVAR")
    if var != "theValue":
        sys.stderr.write(
            "error: ENVAR environment variable must be set to 'theValue'.\n"
        )
        sys.exit(1)


def test_cwd_envVar(results):
    if not results.no_opts:
        homeDir = os.getenv("HOME")
        if homeDir != os.getenv("PWD"):
            sys.stderr.write(
                "error: HOME environment variable must be set to the current"
                " working directory.\n"
            )
            sys.exit(1)


def test_config_file(results):
    if results.config_file:
        with open(results.config_file) as co:
            config_string = co.read()
            if (
                config_string
                != "# This is a demo configuration file\nnumInput=4\nstrInput='coin;plop'"
            ):
                sys.stderr.write(
                    f"error: invalid configuration file:\n {config_string}\n"
                )
                sys.exit(1)


def test_conditional_output_paths(results):
    if results.cond_out:
        path = f"{results.output_file}_default.txt"
        if results.string and results.number > 10:
            path = f"{results.output_file}_{results.string}_{results.number}.txt"
        elif results.string:
            path = f"{results.output_file}_{results.string}.txt"

        with open(results.output_file + "_out1.txt", "w") as f:
            f.write("File content")
        stripped_extensions = [".one", ".two", ".three", ".four", ".five"]
        for extension in stripped_extensions:
            path = path.replace(extension, "")
        with open(path, "w") as f:
            f.write("File content")
    else:
        with open(results.output_file, "w") as f:
            f.write("File content")


def main(args=None):
    parser = argparse.ArgumentParser(
        description="A tool to test Boutiques",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument(
        "-c",
        "--config_file",
        type=lambda x: file_exists(parser, x),
        help="A config file with a number.",
    )
    parser.add_argument("-i", "--string_input", nargs="+", help="One or more strings.")
    parser.add_argument(
        "-l", "--int_input", nargs="+", help="One or more ints.", type=int
    )
    parser.add_argument(
        "-t",
        "--file_list",
        nargs="+",
        help="One or more files.",
        type=lambda x: file_exists(parser, x),
    )
    parser.add_argument(
        "file",
        metavar="file_input",
        type=lambda x: file_exists(parser, x),
        nargs=1,
        help="Any existing file.",
    )
    parser.add_argument("-f", "--flag", action="store_true", help="A flag.")
    parser.add_argument(
        "-n",
        "--number",
        type=int,
        help="A number. Cannot be used when enum_input is used",
    )
    parser.add_argument("-s", "--string", help="A string")
    parser.add_argument(
        "-e",
        "--enum_input",
        type=lambda x: is_valid_enum_value(parser, x),
        nargs=1,
        help='A string in {"val1", "val2", "val3"}.',
    )
    parser.add_argument("output_file", help="The output file.")
    parser.add_argument(
        "--no-opts", action="store_true", help="Ignore container options."
    )
    parser.add_argument(
        "--cond_out",
        action="store_true",
        help="Create conditional output Files",
    )

    results = parser.parse_args() if args is None else parser.parse_args(args)

    test_mutex_group(results)
    test_one_is_required_group(results)
    test_envVar(results)
    test_cwd_envVar(results)
    test_config_file(results)
    sys.stdout.write("This is stdout")
    sys.stderr.write("This is stderr")
    test_conditional_output_paths(results)


if __name__ == "__main__":
    main()



"""
Sample Docopt script for bosh import testing

Usage:
  test_import X Y Z [options]

Options:
  -a --ay         A
  -b --bee        B
  -c --cee        C
  -h --help       Help descriptions
  -v --version    Versions
"""



#!/usr/bin/env python


def raise_error(error_class, message, r=None):
    if r is not None:
        raise error_class(f"[ ERROR ({r.status_code}) ] {message}")
    else:
        raise error_class(f"[ ERROR ] {message}")


def print_info(message, r=None):
    if r is not None:
        print(f"[ INFO ({r.status_code}) ] {message}")
    else:
        print(f"[ INFO ] {message}")


def print_warning(message):
    print(f"[ WARNING ] {message}")


def print_error(message):
    print(f"[ ERROR ] {message}")



#!/usr/bin/env python

import subprocess

import pytest

from boutiques.bosh import bosh
from boutiques.tests.BaseTest import BaseTest
from boutiques.validator import DescriptorValidationError


class TestValidator(BaseTest):
    @pytest.fixture(autouse=True)
    def set_test_dir(self):
        self.setup("validator")

    def test_success(self):
        self.setup("invocation")
        fil = self.get_file_path("good.json")
        self.assertIsNone(bosh(["validate", "--format", fil]))

    def test_success_cli(self):
        self.setup("invocation")
        fil = self.get_file_path("good.json")
        command = "bosh validate " + fil
        process = subprocess.Popen(
            command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        stdout = process.stdout.read().decode("utf-8").strip()
        self.assertEqual(stdout, "OK")
        self.assertFalse(process.returncode)

    def test_fail(self):
        fil = self.get_file_path("bad.json")
        self.assertRaises(DescriptorValidationError, bosh, ["validate", fil])

    def test_invalid_boutiques(self):
        fil = self.get_file_path("invalid.json")
        self.assertRaises(DescriptorValidationError, bosh, ["validate", fil])

    @pytest.mark.skip(reason="Fails after transition to GH action")
    def test_invalid_json_debug(self):
        fil = self.get_file_path("invalid_json.json")
        command = "bosh validate " + fil
        process = subprocess.Popen(command, shell=True, stderr=subprocess.PIPE)
        stderr = process.stderr.read()[-59:].decode("utf-8").strip()
        self.assertEqual(
            stderr,
            "Expecting ',' delimiter or '}': line 9" + " column 2 (char 243)",
        )

    def test_invalid_json(self):
        fil = self.get_file_path("test_exclusive_minimum.json")
        self.assertRaises(DescriptorValidationError, bosh, ["validate", fil])

    def test_invalid_cli(self):
        fil = self.get_file_path("invalid.json")
        command = "bosh validate " + fil
        process = subprocess.Popen(
            command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
        )
        process.communicate()
        self.assertTrue(process.returncode)

    def test_invalid_groups(self):
        fil = self.get_file_path("invalid_groups.json")
        self.assertRaises(DescriptorValidationError, bosh, ["validate", fil])

    def test_invalid_container_index(self):
        fil = self.get_file_path("test_conIndexImage.json")
        command = "bosh validate " + fil
        process = subprocess.Popen(command, shell=True, stdout=subprocess.PIPE)
        stdout = process.stdout.read().decode("utf-8").strip()
        self.assertEqual(
            stdout,
            (
                "ContainerError: container image"
                ' "docker://index.docker.io" is prepended by index'
                " that doesn't match container index value"
                ' "shub://"'
            ),
        )



#!/usr/bin/env python

import os

import pytest

import boutiques as bosh
from boutiques import __file__ as bfile
from boutiques.localExec import ExecutorError
from boutiques.tests.BaseTest import BaseTest


class TestExec(BaseTest):
    @pytest.fixture(autouse=True)
    def set_test_dir(self):
        self.setup("exec")

    def test_failing_launch(self):
        self.assertRaises(
            ExecutorError,
            bosh.execute,
            (
                "launch",
                self.get_file_path("fake.json"),
                self.get_file_path("test_baremetal_invoc.json"),
                "--skip-data-collection",
            ),
        )
        self.assertRaises(
            ExecutorError,
            bosh.execute,
            (
                "launch",
                self.get_file_path("test_baremetal.json"),
                self.get_file_path("fake.json"),
                "--skip-data-collection",
            ),
        )
        self.assertRaises(
            ExecutorError,
            bosh.execute,
            (
                "launch",
                self.get_file_path("test_baremetal.json"),
                self.get_file_path("../boutiques_mocks.py"),
                "--skip-data-collection",
            ),
        )

    def test_no_container(self):
        self.assertFalse(
            bosh.execute(
                "launch",
                self.get_file_path("no_container.json"),
                self.get_file_path("no_container_invocation.json"),
                "--skip-data-collection",
            ).exit_code
        )

    def test_bare_metal_execution(self):
        e = bosh.execute(
            "launch",
            "--no-container",
            self.get_file_path("test_baremetal.json"),
            self.get_file_path("test_baremetal_invoc.json"),
        )
        stdout = e.stdout
        if os.path.exists("test_baremetal_exec.txt"):
            os.remove("test_baremetal_exec.txt")
        self.assertEqual(stdout, "Bare metal execution\n")



from unittest import mock

from boutiques_mocks import MockZenodoRecord, mock_zenodo_search

from boutiques.bosh import bosh
from boutiques.tests.BaseTest import BaseTest


def mock_get(query, exact, *args, include_version=True, **kwargs):
    max_results = args[0].split("=")[-1]

    # Long description text to test truncation
    long_description = (
        "Lorem ipsum dolor sit amet, consectetur adipiscing "
        "elit, sed do eiusmod tempor incididunt ut labore et "
        "dolore magna aliqua. Ut enim ad minim veniam, quis "
        "nostrud exercitation ullamco laboris nisi ut aliquip "
        "ex ea commodo consequat."
    )

    mock_records = []
    # Return an arbitrary list of results with length max_results
    if query == "boutiques":
        for i in range(0, int(max_results)):
            mock_records.append(
                MockZenodoRecord(
                    i,
                    f"Example Tool {i}",
                    long_description,
                    f"exampleTool{i}.json",
                    i,
                )
            )
    # Return only records containing the query
    else:
        mock_records.append(MockZenodoRecord(1234567, query))
        if not exact:
            mock_records.append(MockZenodoRecord(1234568, "foo-" + query))
            mock_records.append(MockZenodoRecord(1234569, query + "-bar"))

    return mock_zenodo_search(mock_records, include_version)


class TestSearch(BaseTest):
    @mock.patch(
        "requests.get",
        side_effect=lambda *args, **kwargs: mock_get("", False, *args, **kwargs),
    )
    def test_search_all(self, mymockget):
        results = bosh(["search"])
        self.assertGreater(len(results), 0)
        self.assertEqual(
            list(results[0].keys()),
            ["ID", "TITLE", "DESCRIPTION", "DOWNLOADS"],
        )

    @mock.patch(
        "requests.get",
        side_effect=lambda *args, **kwargs: mock_get(
            "Example Tool 5", False, *args, **kwargs
        ),
    )
    def test_search_query(self, mymockget):
        results = bosh(["search", "Example Tool 5"])
        self.assertGreater(len(results), 0)
        self.assertIn("Example Tool 5", [d["TITLE"] for d in results])
        self.assertIn("foo-Example Tool 5", [d["TITLE"] for d in results])
        self.assertIn("Example Tool 5-bar", [d["TITLE"] for d in results])

    @mock.patch(
        "requests.get",
        side_effect=lambda *args, **kwargs: mock_get(
            "Example Tool 5", True, *args, **kwargs
        ),
    )
    def test_search_exact_match(self, mymockget):
        results = bosh(["search", "Example Tool 5", "--exact"])
        self.assertGreater(len(results), 0)
        self.assertIn("Example Tool 5", [d["TITLE"] for d in results])
        self.assertNotIn("foo-Example Tool 5", [d["TITLE"] for d in results])
        self.assertNotIn("Example Tool 5-bar", [d["TITLE"] for d in results])

    @mock.patch(
        "requests.get",
        side_effect=lambda *args, **kwargs: mock_get("", False, *args, **kwargs),
    )
    def test_search_verbose(self, mymockget):
        results = bosh(["search", "-v"])
        self.assertGreater(len(results), 0)
        self.assertEqual(
            list(results[0].keys()),
            [
                "ID",
                "TITLE",
                "DESCRIPTION",
                "PUBLICATION DATE",
                "DEPRECATED",
                "DOWNLOADS",
                "AUTHOR",
                "VERSION",
                "DOI",
                "SCHEMA VERSION",
                "CONTAINER",
                "TAGS",
            ],
        )

    @mock.patch(
        "requests.get",
        side_effect=lambda *args, **kwargs: mock_get(
            "boutiques", False, *args, **kwargs
        ),
    )
    def test_search_verbose_missing_metadata(self, mymockget):
        mymockget.side_effect = lambda *args, **kwargs: mock_get(
            "", False, *args, include_version=False, **kwargs
        )
        results = bosh(["search", "-v"])
        self.assertGreater(len(results), 0)
        self.assertEqual(
            list(results[0].keys()),
            [
                "ID",
                "TITLE",
                "DESCRIPTION",
                "PUBLICATION DATE",
                "DEPRECATED",
                "DOWNLOADS",
                "AUTHOR",
                "VERSION",
                "DOI",
                "SCHEMA VERSION",
                "CONTAINER",
                "TAGS",
            ],
        )
        self.assertEqual(results[0]["VERSION"], "unknown")

    @mock.patch(
        "requests.get",
        side_effect=lambda *args, **kwargs: mock_get(
            "boutiques", False, *args, **kwargs
        ),
    )
    def test_search_specify_max_results(self, mymockget):
        results = bosh(["search", "-m", "20"])
        self.assertEqual(len(results), 20)

    @mock.patch(
        "requests.get",
        side_effect=lambda *args, **kwargs: mock_get(
            "boutiques", False, *args, **kwargs
        ),
    )
    def test_search_sorts_by_num_downloads(self, mymockget):
        results = bosh(["search"])
        downloads = []
        for r in results:
            downloads.append(r["DOWNLOADS"])
        self.assertEqual(sorted(downloads, reverse=True), downloads)

    @mock.patch(
        "requests.get",
        side_effect=lambda *args, **kwargs: mock_get(
            "boutiques", False, *args, **kwargs
        ),
    )
    def test_search_truncates_long_text(self, mymockget):
        results = bosh(["search"])
        for r in results:
            for k, v in r.items():
                self.assertLessEqual(len(str(v)), 43)

    @mock.patch(
        "requests.get",
        side_effect=lambda *args, **kwargs: mock_get(
            "boutiques", False, *args, **kwargs
        ),
    )
    def test_search_no_trunc(self, mymockget):
        results = bosh(["search", "--no-trunc"])
        has_no_trunc = False
        for r in results:
            for k, v in r.items():
                if len(str(v)) > 43:
                    has_no_trunc = True
                    break
            else:
                continue
            break
        self.assertTrue(has_no_trunc)
